"""Module to support E&J Technology BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from enum import IntEnum
from string import hexdigits
from typing import Final, Literal

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS


class Cmd(IntEnum):
    """BMS operation codes."""

    RT = 0x2
    CAP = 0x10


class BMS(BaseBMS):
    """E&J Technology BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "E&J Technology",
        "default_model": "smart BMS",
    }
    _BT_MODULE_MSG: Final[bytes] = bytes([0x41, 0x54, 0x0D, 0x0A])  # BLE module message
    _IGNORE_CRC: Final[str] = "libattU"
    _HEAD: Final[bytes] = b"\x3a"
    _TAIL: Final[bytes] = b"\x7e"
    _MAX_CELLS: Final[int] = 16
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp(
            "current", 89, 8, False, lambda x: ((x >> 16) - (x & 0xFFFF)) / 100, Cmd.RT
        ),
        BMSDp("battery_level", 123, 2, False, lambda x: x, Cmd.RT),
        BMSDp("cycle_charge", 15, 4, False, lambda x: x / 10, Cmd.CAP),
        BMSDp(
            "temperature", 97, 2, False, lambda x: x - 40, Cmd.RT
        ),  # only 1st sensor relevant
        BMSDp("cycles", 115, 4, False, lambda x: x, Cmd.RT),
        BMSDp(
            "problem_code", 105, 4, False, lambda x: x & 0x0FFC, Cmd.RT
        ),  # mask status bits
        BMSDp("dischrg_mosfet", 105, 1, False, lambda x: bool(x & 0x1), Cmd.RT),
        BMSDp("chrg_mosfet", 105, 1, False, lambda x: bool(x & 0x2), Cmd.RT),
        BMSDp("balancer", 111, 4, False, lambda x: x, Cmd.RT)
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._data_final: bytearray = bytearray()

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return (
            [  # Lithtech Energy (2x), Volthium
                MatcherPattern(local_name=pattern, connectable=True)
                for pattern in ("L-12V???AH-*", "LT-12V-*", "V-12V???Ah-*")
            ]
            + [  # Fliteboard, Electronix battery
                {
                    "local_name": "libatt*",
                    "manufacturer_id": 21320,
                    "connectable": True,
                },
                {"local_name": "SV12V*", "manufacturer_id": 33384, "connectable": True},
            ]
            + [  # LiTime
                MatcherPattern(  # LiTime based on ser#
                    local_name="LT-12???BG-A0[0-6]*",
                    manufacturer_id=m_id,
                    connectable=True,
                )
                for m_id in (33384, 22618)
            ]
            + [  # LiTime based on ser#
                {
                    "local_name": "LT-24???B-A00[0-2]*",
                    "manufacturer_id": 22618,
                    "connectable": True,
                }
            ]
        )

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return ["6e400001-b5a3-f393-e0a9-e50e24dcca9e"]

    @staticmethod
    def uuid_rx() -> str:
        """Return 128-bit UUID of characteristic that provides notification/read property."""
        return "6e400003-b5a3-f393-e0a9-e50e24dcca9e"

    @staticmethod
    def uuid_tx() -> str:
        """Return 128-bit UUID of characteristic that provides write property."""
        return "6e400002-b5a3-f393-e0a9-e50e24dcca9e"

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {
                "battery_charging",
                "cycle_capacity",
                "delta_voltage",
                "power",
                "runtime",
                "voltage",
            }
        )  # calculate further values from BMS provided set ones

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""

        if data.startswith(BMS._BT_MODULE_MSG):
            self._log.debug("filtering AT cmd")
            if not (data := data.removeprefix(BMS._BT_MODULE_MSG)):
                return

        if data.startswith(BMS._HEAD):  # check for beginning of frame
            self._data.clear()

        self._data += data

        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._data else "cnt.", data
        )

        exp_frame_len: Final[int] = (
            int(self._data[7:11], 16)
            if len(self._data) > 10
            and all(chr(c) in hexdigits for c in self._data[7:11])
            else 0xFFFF
        )

        if not self._data.startswith(BMS._HEAD) or (
            not self._data.endswith(BMS._TAIL) and len(self._data) < exp_frame_len
        ):
            return

        if not self._data.endswith(BMS._TAIL):
            self._log.debug("incorrect EOF: %s", data)
            self._data.clear()
            return

        if not all(chr(c) in hexdigits for c in self._data[1:-1]):
            self._log.debug("incorrect frame encoding.")
            self._data.clear()
            return

        if len(self._data) != exp_frame_len:
            self._log.debug(
                "incorrect frame length %i != %i",
                len(self._data),
                exp_frame_len,
            )
            self._data.clear()
            return

        if not self.name.startswith(BMS._IGNORE_CRC) and (
            crc := BMS._crc(self._data[1:-3])
        ) != int(self._data[-3:-1], 16):
            # libattU firmware uses no CRC, so we ignore it
            self._log.debug(
                "invalid checksum 0x%X != 0x%X", int(self._data[-3:-1], 16), crc
            )
            self._data.clear()
            return

        self._log.debug(
            "address: 0x%X, command 0x%X, version: 0x%X, length: 0x%X",
            int(self._data[1:3], 16),
            int(self._data[3:5], 16) & 0x7F,
            int(self._data[5:7], 16),
            len(self._data),
        )
        self._data_final = self._data.copy()
        self._data_event.set()

    @staticmethod
    def _crc(data: bytearray) -> int:
        return (sum(data) ^ 0xFF) & 0xFF

    @staticmethod
    def _cell_voltages(
        data: bytearray,
        *,
        cells: int,
        start: int,
        size: int = 2,
        gap: int = 0,
        byteorder: Literal["little", "big"] = "big",
        divider: int = 1000,
    ) -> list[float]:
        """Return cell voltages from status message."""
        return [
            (value / divider)
            for idx in range(cells)
            if (value := int(data[start + size * idx : start + size * (idx + 1)], 16))
        ]

    @staticmethod
    def _conv_data(data: dict[int, bytearray]) -> BMSSample:
        result: BMSSample = {}
        for field in BMS._FIELDS:
            result[field.key] = field.fct(
                int(data[field.idx][field.pos : field.pos + field.size], 16)
            )
        return result

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        raw_data: dict[int, bytearray] = {}

        # query real-time information and capacity
        for cmd in (b":000250000E03~", b":001031000E05~"):
            await self._await_reply(cmd)
            rsp: int = int(self._data_final[3:5], 16) & 0x7F
            raw_data[rsp] = self._data_final
            if rsp == Cmd.RT and len(self._data_final) == 0x8C:
                # handle metrisun version
                self._log.debug("single frame protocol detected")
                raw_data[Cmd.CAP] = bytearray(15) + self._data_final[125:]
                break

        if len(raw_data) != len(list(Cmd)) or not all(
            len(value) > 0 for value in raw_data.values()
        ):
            return {}

        return self._conv_data(raw_data) | {
            "cell_voltages": BMS._cell_voltages(
                raw_data[Cmd.RT], cells=BMS._MAX_CELLS, start=25, size=4
            )
        }



"""Test the BLE Battery Management System base class functions."""

from collections.abc import Buffer, Callable
from logging import DEBUG
from typing import Final, NoReturn
from uuid import UUID

from bleak.assigned_numbers import CharacteristicPropertyName
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.backends.service import BleakGATTServiceCollection
from bleak.exc import BleakDeviceNotFoundError, BleakError
from bleak.uuids import normalize_uuid_str
import pytest

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import (
    BaseBMS,
    barr2str,
    crc8,
    crc_modbus,
    crc_sum,
    crc_xmodem,
    lrc_modbus,
)
from aiobmsble.bms.dummy_bms import BMS as DummyBMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient


class MockWriteModeBleakClient(MockBleakClient):
    """Emulate a BleakClient with selectable write mode response."""

    # The following attributes are used to simulate the behavior of the BleakClient
    # They need to be set via monkeypatching in the test since init() is called by the BMS
    PATTERN: list[bytes | Exception | None] = []  # data that is set to problem_code
    VALID_WRITE_MODES: list[CharacteristicPropertyName] = [
        "write-without-response",
        "write",
    ]
    EXP_WRITE_RESPONSE: list[bool] = []

    async def write_gatt_char(
        self,
        char_specifier: BleakGATTCharacteristic | int | str | UUID,
        data: Buffer,
        response: bool | None = None,
    ) -> None:
        """Issue write command to GATT."""
        await super().write_gatt_char(char_specifier, data, response)

        assert self._notify_callback is not None
        if self.PATTERN:
            # check if we have a pattern to return
            pattern: bytes | Exception | None = self.PATTERN.pop(0)
            exp_wr_mode: Final[bool] = self.EXP_WRITE_RESPONSE.pop(0)
            if isinstance(pattern, Exception):
                raise pattern

            req_wr_mode: Final[str] = "write" if response else "write-without-response"
            assert response == exp_wr_mode, "write response mismatch"

            if isinstance(pattern, bytes) and req_wr_mode in self.VALID_WRITE_MODES:
                # check if we have a dict to return
                self._notify_callback("rx_char", bytearray(pattern))
                return

            # if None was selected do not return (trigger timeout) and wait for next pattern
            return

        # no pattern left, raise exception
        raise ValueError


class MinTestBMS(BaseBMS):
    """Minimal Test BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Test Manufacturer",
        "default_model": "minimal BMS for test",
    }

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [{"local_name": "Test", "connectable": True}]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("afe0")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "afe1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "afe2"

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)
        # do not set event to make tests fail if wait_for_notify is not set

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_reply(b"mock_command", wait_for_notify=False)  # do not wait
        return {"problem_code": 21}


class DataTestBMS(MinTestBMS):
    """BMS providing simple data to test, e.g. value calculation."""

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset({"power", "battery_charging"})

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_reply(b"mock_command", wait_for_notify=False)  # do not wait
        return {
            "voltage": 13,
            "current": 1.7,
            "cycle_charge": 19,
            "cycles": 23,
            "problem_code": 21,
        }


class WMTestBMS(MinTestBMS):
    """Write mode mock BMS implementation."""

    def __init__(
        self,
        char_tx_properties: list[str],
        ble_device: BLEDevice,
        keep_alive: bool = True,
    ) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._char_tx_properties: list[str] = char_tx_properties

    def _wr_response(self, char: int | str) -> bool:
        return bool("write" in self._char_tx_properties)

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)
        self._data = data
        self._data_event.set()

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_reply(b"mock_command")

        return {"problem_code": int.from_bytes(self._data, "big", signed=False)}


@pytest.mark.parametrize(
    ("bt_patch", "result_patch"),
    [
        ({"2a28": b"mock_SW_version"}, {"sw_version": "mock_SW_version"}),
        ({"2a28": b""}, {}),
        ({}, {}),
    ],
    ids=["defaults", "empty", "no_char"],
)
async def test_device_info(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    bt_patch,
    result_patch,
) -> None:
    """Verify that device_info reads BLE characteristic 180A and provides default values."""
    defaults = MockBleakClient.BT_INFO.copy()
    del defaults["2a28"]
    monkeypatch.setattr(MockBleakClient, "BT_INFO", defaults | bt_patch)
    patch_bleak_client()
    bms: MinTestBMS = MinTestBMS(generate_ble_device())
    assert (
        await bms.device_info()
        == {
            "fw_version": "mock_FW_version",
            "model": "mock_model",
            "serial_number": "mock_serial_number",
            "hw_version": "mock_HW_version",
            "manufacturer": "mock_manufacturer",
        }
        | result_patch
    )


async def test_device_info_fail(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client: Callable[..., None]
) -> None:
    """Test only BMS default information is returned if characteristic 0x180a does not exit."""
    monkeypatch.setattr(
        BleakGATTServiceCollection, "get_service", lambda obj, char: None
    )
    patch_bleak_client()
    bms: MinTestBMS = MinTestBMS(generate_ble_device())
    await bms.async_update()  # run update to have connection open
    assert not await bms.device_info()  # if characteristic does not exist, no ouput
    assert bms.name == "MockBLEDevice"  # name is gathered from BLEDevice
    assert bms._client.is_connected


def test_calc_missing_values(bms_data_fixture: BMSSample) -> None:
    """Check if missing data is correctly calculated."""
    bms_data: BMSSample = bms_data_fixture
    ref: BMSSample = bms_data_fixture.copy()

    BaseBMS._add_missing_values(
        bms_data,
        frozenset(
            {
                "battery_charging",
                "cycle_capacity",
                "power",
                "runtime",
                "delta_voltage",
                "temperature",
                "voltage",  # check value is not overwritten
            }
        ),
    )
    ref = ref | {
        "cycle_capacity": 238,
        "delta_voltage": 0.111,
        "power": (
            -91
            if bms_data.get("current", 0) < 0
            else 0 if bms_data.get("current") == 0 else 147
        ),
        # battery is charging if current is positive
        "battery_charging": bms_data.get("current", 0) > 0,
        "temperature": -34.396,
        "problem": False,
    }
    if bms_data.get("current", 0) < 0:
        ref |= {"runtime": 9415}

    assert bms_data == ref


def test_calc_voltage() -> None:
    """Check if missing data is correctly calculated."""
    bms_data: BMSSample = {"cell_voltages": [3.456, 3.567]}
    ref: BMSSample = bms_data.copy()
    BaseBMS._add_missing_values(bms_data, frozenset({"voltage"}))
    assert bms_data == ref | {"voltage": 7.023, "problem": False}


def test_calc_cycle_chrg() -> None:
    """Check if missing data is correctly calculated."""
    bms_data: BMSSample = {"battery_level": 73, "design_capacity": 125}
    ref: BMSSample = bms_data.copy()
    BaseBMS._add_missing_values(bms_data, frozenset({"cycle_charge"}))
    assert bms_data == ref | {"cycle_charge": 91.25, "problem": False}


def test_calc_battery_level() -> None:
    """Check if missing battery_level is correctly calculated."""
    bms_data: BMSSample = {"cycle_charge": 421, "design_capacity": 983}
    ref: BMSSample = bms_data.copy()
    BaseBMS._add_missing_values(bms_data, frozenset({"battery_level"}))
    assert bms_data == ref | {"battery_level": 42.8, "problem": False}


def test_calc_cycles() -> None:
    """Check if missing cycle is correctly calculated."""
    bms_data: BMSSample = {"total_charge": 1234567, "design_capacity": 256}
    ref: BMSSample = bms_data.copy()
    BaseBMS._add_missing_values(bms_data, frozenset({"cycles"}))
    assert bms_data == ref | {"cycles": 4822, "problem": False}


@pytest.mark.parametrize(
    ("raw", "expected"),
    [(True, {}), (False, {"battery_charging": True, "power": 22.1, "problem": True})],
    ids=["raw", "ext"],
)
async def test_async_update(
    patch_bleak_client: Callable[..., None], raw: bool, expected: dict
) -> None:
    """Check update function of the BMS returns calues."""
    patch_bleak_client()
    bms: DataTestBMS = DataTestBMS(generate_ble_device())
    assert await bms.async_update(raw=raw) == BMSSample(
        {
            "voltage": 13,
            "current": 1.7,
            "cycle_charge": 19,
            "cycles": 23,
            "problem_code": 21,
        }
    ) | BMSSample(**expected)


@pytest.mark.parametrize(
    ("problem_sample"),
    [
        ({"voltage": -1}, "negative overall voltage"),
        ({"cell_voltages": [5.907]}, "high cell voltage"),
        ({"cell_voltages": [-0.001]}, "negative cell voltage"),
        ({"delta_voltage": 5.907}, "doubtful delta voltage"),
        ({"cycle_charge": 0}, "doubtful cycle charge"),
        ({"battery_level": 101}, "doubtful SoC"),
        ({"problem_code": 0x1}, "BMS problem code"),
        ({"problem": True}, "BMS problem report"),
    ],
    ids=lambda param: param[1],
)
def test_problems(problem_sample: tuple[BMSSample, str]) -> None:
    """Check if missing data is correctly calculated."""
    bms_data: BMSSample = problem_sample[0].copy()

    BaseBMS._add_missing_values(bms_data, frozenset({"runtime"}))

    assert bms_data == problem_sample[0] | {"problem": True}


@pytest.mark.parametrize(
    ("replies", "exp_wr_response", "exp_output"),
    [
        ([b"\x12"], [True], [0x12]),
        (
            [None] * 2 * (BaseBMS.MAX_RETRY),
            [True] * (BaseBMS.MAX_RETRY) + [False] * (BaseBMS.MAX_RETRY),
            [TimeoutError()],
        ),
        (
            [None] * (BaseBMS.MAX_RETRY - 1) + [b"\x13"],
            [True] * (BaseBMS.MAX_RETRY),
            [0x13],
        ),
        (
            [None] * (BaseBMS.MAX_RETRY) + [b"\x14"],
            [True] * (BaseBMS.MAX_RETRY) + [False],
            [0x14],
        ),
        (
            [BleakError()]
            + [None] * (BaseBMS.MAX_RETRY - 1)
            + [b"\x15"]
            + [None] * (BaseBMS.MAX_RETRY - 1)
            + [b"\x16"],
            [True] + [False] * BaseBMS.MAX_RETRY + [False] * BaseBMS.MAX_RETRY,
            [0x15, 0x16],
        ),
        (
            [BleakDeviceNotFoundError("mock_device")],
            [False],
            [BleakDeviceNotFoundError("mock_device")],
        ),
        (
            [None] * (BaseBMS.MAX_RETRY - 1) + [ValueError()],
            [True] * (BaseBMS.MAX_RETRY),
            [ValueError()],
        ),
    ],
    ids=[
        "basic_test",
        "no_response",
        "retry_count-1",
        "retry_count",
        "mode_switch",
        "no-retry-exc",
        "unhandled_exc",
    ],
)
async def test_write_mode(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    patch_bms_timeout: Callable[..., None],
    replies: list[bytearray | Exception | None],
    exp_wr_response: list[bool],
    exp_output: list[int | Exception],
    request: pytest.FixtureRequest,
) -> None:
    """Check if write mode selection works correctly."""

    assert len(replies) == len(
        exp_wr_response
    ), "Replies and expected responses must match in length!"
    patch_bms_timeout()
    monkeypatch.setattr(MockWriteModeBleakClient, "PATTERN", replies)
    monkeypatch.setattr(MockWriteModeBleakClient, "EXP_WRITE_RESPONSE", exp_wr_response)

    patch_bleak_client(MockWriteModeBleakClient)

    bms = WMTestBMS(
        ["write-no-response", "write"],
        generate_ble_device(),
        False,
    )

    # NOTE: output must reflect the end result after one call, as init of HA resets the whole BMS!
    for output in exp_output:
        if isinstance(output, Exception):
            with pytest.raises(type(output)):
                await bms.async_update()
        else:
            assert await bms.async_update() == {
                "problem_code": output
            }, f"{request.node.name} failed!"


async def test_wr_mode_reset(
    monkeypatch: pytest.MonkeyPatch, patch_bleak_client: Callable[..., None]
) -> None:
    """Check that write mode selection is reset on disconnect of the BMS."""
    monkeypatch.setattr(MockWriteModeBleakClient, "PATTERN", [b"\x42"])
    monkeypatch.setattr(MockWriteModeBleakClient, "EXP_WRITE_RESPONSE", [False])
    patch_bleak_client(MockWriteModeBleakClient)

    bms: WMTestBMS = WMTestBMS(["write_without"], generate_ble_device())
    assert await bms.async_update() == {"problem_code": 0x42}
    assert bms._inv_wr_mode is False
    await bms.disconnect(True)
    assert bms._inv_wr_mode is None


def test_get_bms_module() -> None:
    """Check that basebms and dummy_bms return correct module name."""
    assert BaseBMS.get_bms_module() == "aiobmsble.basebms"
    assert DummyBMS.get_bms_module() == "aiobmsble.bms.dummy_bms"


async def test_no_notify(
    patch_bleak_client: Callable[..., None], caplog: pytest.LogCaptureFixture
) -> None:
    """Test BMS update without waiting for notification event."""
    patch_bleak_client(MockBleakClient)

    bms: MinTestBMS = MinTestBMS(generate_ble_device(), keep_alive=False)
    with caplog.at_level(DEBUG):
        result: BMSSample = await bms.async_update()
    assert "MockBleakClient write_gatt_char afe2, data: b'mock_command'" in caplog.text
    assert result == {"problem_code": 21}
    assert not bms._client.is_connected


async def test_disconnect_fail(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Check that exceptions in connect function for guarding disconnect are ignored."""

    async def _raise_bleak_error(_args) -> NoReturn:
        raise BleakError

    monkeypatch.setattr(MockBleakClient, "disconnect", _raise_bleak_error)
    patch_bleak_client(MockBleakClient)

    bms: MinTestBMS = MinTestBMS(generate_ble_device(), keep_alive=False)
    with caplog.at_level(DEBUG):
        result: BMSSample = await bms.async_update()
    assert result == {"problem_code": 21}
    assert "failed to disconnect stale connection (BleakError)" in caplog.text
    assert "disconnect failed!" in caplog.text


async def test_init_connect_fail(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client: Callable[..., None],
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Check that exceptions in connect function for guarding disconnect are ignored."""

    async def _raise_value_error(_args) -> NoReturn:
        raise ValueError("MockValueError")

    patch_bleak_client(MockBleakClient)
    monkeypatch.setattr(MinTestBMS, "_init_connection", _raise_value_error)

    bms: MinTestBMS = MinTestBMS(generate_ble_device())
    with caplog.at_level(DEBUG), pytest.raises(ValueError, match="MockValueError"):
        await bms.async_update()


async def test_context_mgr(
    patch_bleak_client: Callable[..., None],
) -> None:
    """Test that context manager provides data."""
    patch_bleak_client(MockBleakClient)

    async with DataTestBMS(generate_ble_device(), keep_alive=True) as bms:
        assert await bms.async_update() == {
            "voltage": 13,
            "current": 1.7,
            "battery_charging": True,
            "power": 22.1,
            "cycle_charge": 19,
            "cycles": 23,
            "problem": True,
            "problem_code": 21,
        }


async def test_context_mgr_fail(
    patch_bleak_client: Callable[..., None],
) -> None:
    """Check that context manager enforces `keep_alive=True`."""

    patch_bleak_client(MockBleakClient)

    with pytest.raises(ValueError, match="usage of context manager*"):
        async with MinTestBMS(generate_ble_device(), keep_alive=False) as bms:
            await bms.async_update()


def test_crc_calculations() -> None:
    """Check if CRC calculations are correct."""
    # Example data for CRC calculation
    data: bytearray = bytearray([0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39])
    test_fn: list[tuple[Callable[[bytearray], int], int]] = [
        (crc_modbus, 0x4B37),
        (crc8, 0xA1),
        (crc_xmodem, 0x31C3),
        (lrc_modbus, 0xFE23),
        (crc_sum, 0xDD),
    ]

    for crc_fn, expected_crc in test_fn:
        calculated_crc: int = crc_fn(data)
        assert (
            calculated_crc == expected_crc
        ), f"Expected {expected_crc}, got {calculated_crc}"


@pytest.mark.parametrize(
    ("data", "cells", "start", "size", "byteorder", "divider", "expected"),
    [
        # Two cells, big endian, default divider
        (bytearray([0x0D, 0x80, 0x0D, 0xF7]), 2, 0, 2, "big", 1000, [3.456, 3.575]),
        # Two cells, little endian, default divider
        (bytearray([0x80, 0x0D, 0xF7, 0x0D]), 2, 0, 2, "little", 1000, [3.456, 3.575]),
        # One cell, big endian, custom divider
        (bytearray([0x30, 0x34]), 1, 0, 2, "big", 10000, [1.234]),
        (  # Three cells, big endian, offset start
            bytearray([0x00, 0x00, 0x0D, 0x80, 0x0D, 0xF7, 0x0D, 0xA7]),
            3,
            2,
            2,
            "big",
            1000,
            [3.456, 3.575, 3.495],
        ),
        # Not enough data for all cells
        (bytearray([0x0D, 0x80]), 2, 0, 2, "big", 1000, [3.456]),
        # Zero cells
        (bytearray([0x0D, 0x80]), 0, 0, 2, "big", 1000, []),
        # Divider = 1 (raw values)
        (bytearray([0x01, 0x02, 0x03, 0x04]), 2, 0, 2, "big", 1, [258, 772]),
    ],
    ids=[
        "two_cells_big_endian",
        "two_cells_little_endian",
        "one_cell_custom_divider",
        "three_cells_offset_start",
        "not_enough_data",
        "zero_cells",
        "divider_one_raw_values",
    ],
)
def test_cell_voltages(data, cells, start, size, byteorder, divider, expected) -> None:
    """Test the _cell_voltages method of BaseBMS with various input parameters."""
    result: list[float] = BaseBMS._cell_voltages(
        data,
        cells=cells,
        start=start,
        size=size,
        byteorder=byteorder,
        divider=divider,
    )
    assert result == expected


@pytest.mark.parametrize(
    (
        "data",
        "values",
        "start",
        "size",
        "byteorder",
        "signed",
        "offset",
        "divider",
        "expected",
    ),
    [
        # Two signed big endian values, no offset, divider=1
        (bytearray([0xFF, 0xEC, 0x00, 0x64]), 2, 0, 2, "big", True, 0, 1, [-20, 100]),
        (  # Two unsigned little endian values, offset=0, divider=10
            bytearray([0x10, 0x00, 0x20, 0x00]),
            2,
            0,
            2,
            "little",
            False,
            0,
            10,
            [1.6, 3.2],
        ),
        # One signed big endian value, offset=273.15, divider=1
        (bytearray([0x0B, 0x98]), 1, 0, 2, "big", False, 2731, 10, [23.7]),
        (  # Three signed little endian values, offset=0, divider=100
            bytearray([0x64, 0x00, 0xC8, 0xFF, 0x2C, 0x01]),
            3,
            0,
            2,
            "little",
            True,
            0,
            100,
            [1.0, -0.56, 3.0],
        ),
        # Not enough data for all values
        (bytearray([0x00, 0x7D]), 2, 0, 2, "big", True, 0, 1, [125]),
        # Zero values requested
        (bytearray([0x00, 0x7D]), 0, 0, 2, "big", True, 0, 1, []),
        # Divider = 1, offset = 7
        (bytearray([0x00, 0x14]), 1, 0, 2, "big", True, 7, 1, [13]),
        # no offset, div = 10
        (
            bytearray(b"\x65\x64\x00\x40"),
            4,
            0,
            1,
            "big",
            True,
            0,
            10,
            [10.1, 10.0, 0.0, 6.4],
        ),
        # offset -40, div = 10
        (
            bytearray(b"\x65\x64\x00\x40"),
            4,
            0,
            1,
            "big",
            True,
            40,
            10,
            [6.1, 6.0, 2.4],
        ),
        # offset -25, div = 1
        (bytearray(b"\x65\x64\x00\x40"), 4, 0, 1, "big", True, 25, 1, [76, 75, 39]),
    ],
    ids=[
        "two_signed_big_endian",
        "two_unsigned_little_endian_div10",
        "one_signed_big_endian_kelvin_offset",
        "three_signed_little_endian_div100",
        "not_enough_data",
        "zero_values",
        "divider1_offset10",
        "no offset",
        "offset-40",
        "div-1",
    ],
)
def test_temp_values(
    data, values, start, size, byteorder, signed, offset, divider, expected
) -> None:
    """Test the _temp_values method of BaseBMS with various input parameters."""
    result: list[int | float] = BaseBMS._temp_values(
        data,
        values=values,
        start=start,
        size=size,
        byteorder=byteorder,
        signed=signed,
        offset=offset,
        divider=divider,
    )
    assert result == expected


@pytest.mark.parametrize(
    ("fields", "data", "byteorder", "offset", "expected"),
    [
        # Test with big endian and multiple data points
        (
            (
                BMSDp("voltage", 0, size=2, signed=False),
                BMSDp("current", 2, size=2, signed=True),
            ),
            bytearray([0x0D, 0x80, 0xFF, 0xEC]),
            "big",
            0,
            {"voltage": 3456.0, "current": -20},
        ),
        # Test with dict data, little endian
        (
            (BMSDp("voltage", 0, size=2, signed=False, fct=lambda x: x / 10, idx=1),),
            {1: bytearray([0x64, 0x00])},
            "little",
            0,
            {"voltage": 10},
        ),
        # Test with missing dict data
        (
            (BMSDp("voltage", 0, size=2, signed=False, idx=2),),
            {1: bytearray([0x64, 0x00])},
            "little",
            0,
            {},
        ),
        # Test with offset
        (
            (BMSDp("battery_level", 1, size=1, signed=False),),
            bytearray([0x00, 0x7D]),
            "big",
            0,
            {"battery_level": 125},
        ),
        # Test with offset shifting the slice
        (
            (BMSDp("voltage", 0, size=2, signed=False),),
            bytearray([0x00, 0x00, 0x12, 0x34]),
            "big",
            2,
            {"voltage": 0x1234},
        ),
    ],
    ids=[
        "be_multiple_dps",
        "le_dict_single_dp_lambda",
        "le_dict_missing_idx",
        "be_with_offset",
        "be_offset_shift_slice",
    ],
)
def test_decode_data(fields, data, byteorder, offset, expected) -> None:
    """Test the _decode_data method of BaseBMS with various input parameters."""
    result: BMSSample = BaseBMS._decode_data(
        fields, data, byteorder=byteorder, offset=offset
    )
    assert result == expected


@pytest.mark.parametrize(
    ("data", "expected"),
    [(b"", ""), (b"\x00 ", ""), (b"test\x00 ", "test"), (b"test  \t\r ", "test")],
    ids=["empty", "hex", "text_hex", "test_space"],
)
def test_barr2str(data: bytes, expected: str) -> None:
    """Test bytearray to string conversion function."""
    assert barr2str(bytearray(data)) == expected



"""Module to support Felicity BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from collections.abc import Callable
from json import JSONDecodeError, loads
from typing import Any, Final, Literal

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS


class BMS(BaseBMS):
    """Felicity BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Felicity Solar",
        "default_model": "LiFePo4 battery",
    }
    _HEAD: Final[bytes] = b"{"
    _TAIL: Final[bytes] = b"}"
    _CMD_PRE: Final[bytes] = b"wifilocalMonitor:"  # CMD prefix
    _CMD_BI: Final[bytes] = b"get dev basice infor"
    _CMD_DT: Final[bytes] = b"get Date"
    _CMD_RT: Final[bytes] = b"get dev real infor"
    _FIELDS: Final[
        list[
            tuple[
                Literal["voltage", "current", "cycle_charge", "battery_level"],
                str,
                Callable[[list[Any]], float],
            ]
        ]
    ] = [
        ("voltage", "Batt", lambda x: x[0][0] / 1000),
        ("current", "Batt", lambda x: x[1][0] / 10),
        (
            "cycle_charge",
            "BatsocList",
            lambda x: (int(x[0][0]) * int(x[0][2])) / 1e7,
        ),
        ("battery_level", "BatsocList", lambda x: x[0][0] / 100),
    ]

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._data_final: dict[str, Any] = {}

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {"local_name": pattern, "connectable": True} for pattern in ("F07*", "F10*")
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("6e6f736a-4643-4d44-8fa9-0fafd005e455")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 128-bit UUID of characteristic that provides notification/read property."""
        return "49535458-8341-43f4-a9d4-ec0e34729bb3"

    @staticmethod
    def uuid_tx() -> str:
        """Return 128-bit UUID of characteristic that provides write property."""
        return "49535258-184d-4bd9-bc61-20c647249616"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        await self._await_reply(BMS._CMD_PRE + BMS._CMD_BI)
        return {
            "fw_version": self._data_final.get("M1SwVer", []),
            "sw_version": self._data_final.get("version", []),
            "model_id": self._data_final.get("Type", []),
            "serial_number": self._data_final.get("DevSN", []),
        }

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {
                "battery_charging",
                "cycle_capacity",
                "delta_voltage",
                "power",
                "runtime",
                "temperature",
            }
        )  # calculate further values from BMS provided set ones

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""

        if data.startswith(BMS._HEAD):
            self._data = bytearray()

        self._data += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._data else "cnt.", data
        )

        if not data.endswith(BMS._TAIL):
            return

        try:
            self._data_final = loads(self._data)
        except (JSONDecodeError, UnicodeDecodeError):
            self._log.debug("JSON decode error: %s", self._data)
            return

        if (ver := self._data_final.get("CommVer", 0)) != 1:
            self._log.debug("Unknown protocol version (%i)", ver)
            return

        self._data_event.set()

    @staticmethod
    def _conv_data(data: dict[str, Any]) -> BMSSample:
        result: BMSSample = {}
        for key, itm, func in BMS._FIELDS:
            result[key] = func(data.get(itm, []))
        return result

    @staticmethod
    def _conv_cells(data: dict[str, Any]) -> list[float]:
        return [(value / 1000) for value in data.get("BatcelList", [])[0]]

    @staticmethod
    def _conv_temp(data: dict[str, Any]) -> list[float]:
        return [
            (value / 10) for value in data.get("BtemList", [])[0] if value != 0x7FFF
        ]

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""

        await self._await_reply(BMS._CMD_PRE + BMS._CMD_RT)

        return (
            BMS._conv_data(self._data_final)
            | {"temp_values": BMS._conv_temp(self._data_final)}
            | {"cell_voltages": BMS._conv_cells(self._data_final)}
            | {
                "problem_code": int(
                    self._data_final.get("Bwarn", 0) + self._data_final.get("Bfault", 0)
                )
            }
        )



"""Test aiobmsble library via fuzzing."""

from asyncio import iscoroutinefunction
from collections.abc import Callable
from types import ModuleType
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.service import BleakGATTService
from hypothesis import given, strategies as st
import pytest

from aiobmsble.basebms import BaseBMS
from tests.bluetooth import generate_ble_device
from tests.conftest import MockBleakClient


@given(
    data=st.binary(min_size=0, max_size=513)
)  # ATT is not allowed larger than 512 bytes
async def test_notification_handler(
    monkeypatch: pytest.MonkeyPatch,
    patch_bleak_client,
    pytestconfig: pytest.Config,
    plugin_fixture: ModuleType,
    data: bytes,
) -> None:
    """Test the notification handler."""

    # fuzzing can run from VScode (no coverage) or command line with option --no-cov
    if {"vscode_pytest"}.issubset(set(pytestconfig.invocation_params.args)) or (
        not pytestconfig.getoption("--no-cov")
    ):
        pytest.skip("Skipping fuzzing tests due to coverage generation!")

    async def patch_init() -> None:
        return

    patch_bleak_client(MockBleakClient)

    bms_instance: BaseBMS = plugin_fixture.BMS(  # char 10 needs to be A|B for OGT
        generate_ble_device(name="MockFuzz-BLE")
    )

    monkeypatch.setattr(
        bms_instance, "_init_connection", patch_init
    )  # required for _init_connection overloads, e.g. JK BMS

    await bms_instance._connect()
    notify_handler: Final[Callable] = bms_instance._notification_handler  # type: ignore[attr-defined]
    notify_characteristics: Final[BleakGATTCharacteristic] = BleakGATTCharacteristic(
        None, 1, "fff4", ["notify"], lambda: 512, BleakGATTService(None, 0, "fff0")
    )

    if iscoroutinefunction(notify_handler):
        await notify_handler(notify_characteristics, bytearray(data))
    else:
        notify_handler(notify_characteristics, bytearray(data))



"""Package for battery management systems (BMS) via Bluetooth LE (aiobmsble).

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from collections.abc import Callable
from enum import IntEnum
from typing import Any, Literal, NamedTuple, TypedDict

type BMSValue = Literal[
    "battery_charging",
    "battery_mode",
    "battery_level",
    # "battery_health",
    "current",
    "power",
    "temperature",
    "voltage",
    "cycles",
    "cycle_capacity",
    "cycle_charge",
    "total_charge",
    "delta_voltage",
    "problem",
    "runtime",
    "balancer",
    "balance_current",
    "cell_count",
    "cell_voltages",
    "design_capacity",
    "pack_count",
    "temp_sensors",
    "temp_values",
    "problem_code",
    "chrg_mosfet",
    "dischrg_mosfet",
    "heater",
]

type BMSpackvalue = Literal[
    "pack_voltages",
    "pack_currents",
    "pack_battery_levels",
    "pack_cycles",
]


class BMSMode(IntEnum):
    """Enumeration of BMS modes."""

    UNKNOWN = -1
    BULK = 0x00
    ABSORPTION = 0x01
    FLOAT = 0x02


class BMSSample(TypedDict, total=False):
    """Dictionary representing a sample of battery management system (BMS) data."""

    battery_charging: bool  # True: battery charging
    battery_mode: BMSMode  # BMS charging mode
    battery_level: int | float  # [%]
    # battery_health: int | float  # [%]
    current: float  # [A] (positive: charging)
    power: float  # [W] (positive: charging)
    temperature: int | float  # [°C]
    voltage: float  # [V]
    cycle_capacity: int | float  # [Wh]
    cycles: int  # [#]
    delta_voltage: float  # [V]
    problem: bool  # True: problem detected
    runtime: int  # [s]

    # detailed information
    balancer: bool | int  # False: off, True: active or bit mask, 1: enabled/active
    balance_current: float  # [A]
    cell_count: int  # [#]
    cell_voltages: list[float]  # [V]
    cycle_charge: int | float  # [Ah]
    total_charge: int  # [Ah], overall discharged
    design_capacity: int  # [Ah]
    pack_count: int  # [#]
    temp_sensors: int  # [#]
    temp_values: list[int | float]  # [°C]
    problem_code: int  # BMS specific code, 0 no problem

    # BMS switches
    chrg_mosfet: bool  # True: enabled
    dischrg_mosfet: bool  # True: enabled
    heater: bool  # True: enabled

    # battery pack data
    pack_voltages: list[float]  # [V]
    pack_currents: list[float]  # [A]
    pack_battery_levels: list[int | float]  # [%]
    pack_cycles: list[int]  # [#]


class BMSDp(NamedTuple):
    """Representation of one BMS data point."""

    key: BMSValue  # the key of the value to be parsed
    pos: int  # position within the message
    size: int  # size in bytes
    signed: bool  # signed value
    fct: Callable[[int], Any] = lambda x: x  # conversion function (default do nothing)
    idx: int = -1  # array index containing the message to be parsed


class BMSInfo(TypedDict, total=False):
    """Human readable information about the BMS device."""

    default_manufacturer: str
    default_model: str
    default_name: str
    fw_version: str
    manufacturer: str
    model: str
    model_id: str
    name: str
    serial_number: str
    sw_version: str
    hw_version: str


class MatcherPattern(TypedDict, total=False):
    """Optional patterns that can match Bleak advertisement data."""

    local_name: str  # name pattern that supports Unix shell-style wildcards
    manufacturer_data_start: list[int]  # start bytes of manufacturer data
    manufacturer_id: int  # required manufacturer ID
    oui: str  # required OUI used in the MAC address (first 3 bytes)
    service_data_uuid: str  # service data for the service UUID
    service_uuid: str  # 128-bit UUID that the device must advertise
    connectable: bool  # True if active connections to the device are required



"""Test the aiobmsble library base class functions."""

from string import hexdigits
from types import ModuleType
from uuid import UUID

from netaddr import OUI, NotRegisteredError

from aiobmsble.basebms import BaseBMS
from aiobmsble.test_data import bms_advertisements, ignore_advertisements
from aiobmsble.utils import bms_supported, load_bms_plugins


def test_bms_id(plugin_fixture: ModuleType) -> None:
    """Test that the BMS returns default information."""
    bms_class: type[BaseBMS] = plugin_fixture.BMS
    for key in ("default_manufacturer", "default_model"):
        assert str(bms_class.INFO.get(key, "")).strip()
    assert len(bms_class.bms_id().strip())


def test_matcher_dict(plugin_fixture: ModuleType) -> None:
    """Test that the BMS returns BT matcher."""
    bms_class: type[BaseBMS] = plugin_fixture.BMS
    assert len(bms_class.matcher_dict_list())
    for matcher in bms_class.matcher_dict_list():

        if manufacturer_id := matcher.get("manufacturer_id"):
            assert (
                manufacturer_id == manufacturer_id & 0xFFFF
            ), f"incorrect {manufacturer_id=}"

        if service_uuid := matcher.get("service_uuid"):
            assert UUID(service_uuid), f"incorrect {service_uuid=}"

        if manufacturer_data_start := matcher.get("manufacturer_data_start"):
            assert all(
                byte == byte & 0xFF for byte in manufacturer_data_start
            ), "manufacturer_data_start needs to contain Byte values!"

        if oui := matcher.get("oui"):
            parts: list[str] = oui.split(":")
            assert len(parts) == 3 and all(
                len(part) == 2 and all(c in hexdigits for c in part) for part in parts
            ), f"incorrect {oui=}"
            try:
                OUI(oui.replace(":", "-"))
            except NotRegisteredError:
                assert (int(parts[0], 16) & 0xC0) not in (
                    0x00,  # Non-resolvable random private address
                    0x40,  # Resolvable random private address
                    # 0x80,  # 	Reserved for future use
                    # 0xC0,  # Static random device address
                ), f"random private address OUI ({oui}) cannot be used for filtering!"


def test_advertisements_unique() -> None:
    """Check that each advertisement only matches one, the right BMS."""
    for adv, mac_addr, bms_real, _comments in bms_advertisements():
        for bms_under_test in load_bms_plugins():
            supported: bool = bms_supported(bms_under_test.BMS, adv, mac_addr)
            assert supported == (
                f"aiobmsble.bms.{bms_real}" == bms_under_test.__name__
            ), f"{adv} {"incorrectly matches"if supported else "does not match"} {bms_under_test}!"


def test_advertisements_ignore() -> None:
    """Check that each advertisement to be ignored is actually ignored."""
    for adv, mac_addr, reason, _comments in ignore_advertisements():
        for bms_under_test in load_bms_plugins():
            supported: bool = bms_supported(bms_under_test.BMS, adv, mac_addr)
            assert (
                not supported
            ), f"{adv} incorrectly matches {bms_under_test}! {reason=}"



"""Module to support CBT Power VB series BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from string import hexdigits
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, lrc_modbus


class BMS(BaseBMS):
    """CBT Power VB series battery class implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Creabest", "default_model": "VB series"}
    _HEAD: Final[bytes] = b"\x7e"
    _TAIL: Final[bytes] = b"\x0d"
    _CMD_VER: Final[int] = 0x11  # TX protocol version
    _RSP_VER: Final[int] = 0x22  # RX protocol version
    _LEN_POS: Final[int] = 9
    _MIN_LEN: Final[int] = _LEN_POS + 3 + len(_HEAD) + len(_TAIL) + 4
    _MAX_LEN: Final[int] = 255
    _CELL_POS: Final[int] = 6

    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 2, 2, False, lambda x: x / 10),
        BMSDp("current", 0, 2, True, lambda x: x / 10),
        BMSDp("battery_level", 4, 2, False, lambda x: min(x, 100)),
        BMSDp("cycles", 7, 2, False),
        BMSDp("problem_code", 15, 6, False, lambda x: x & 0xFFF000FF000F),
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._exp_len: int = 0

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {  # Creabest
                "service_uuid": normalize_uuid_str("fff0"),
                "manufacturer_id": 16963,
                "connectable": True,
            },
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [
            normalize_uuid_str("ffe0"),
            normalize_uuid_str("ffe5"),
        ]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ffe4"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ffe9"

    # async def _fetch_device_info(self) -> BMSInfo: unknown, use default

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {
                "battery_charging",
                "delta_voltage",
                "temperature",
                "power",
                "runtime",
                "cycle_capacity",
                "cycle_charge",
            }
        )  # calculate further values from BMS provided set ones

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""

        if len(data) > BMS._LEN_POS + 4 and data.startswith(BMS._HEAD):
            self._data = bytearray()
            try:
                length: Final[int] = int(data[BMS._LEN_POS : BMS._LEN_POS + 4], 16)
                self._exp_len = length & 0xFFF
                if BMS.lencs(length) != length >> 12:
                    self._exp_len = 0
                    self._log.debug("incorrect length checksum.")
            except ValueError:
                self._exp_len = 0

        self._data += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._data else "cnt.", data
        )

        if len(self._data) < self._exp_len + BMS._MIN_LEN:
            return

        if not self._data.endswith(BMS._TAIL):
            self._log.debug("incorrect EOF: %s", data)
            self._data.clear()
            return

        if not all(chr(c) in hexdigits for c in self._data[1:-1]):
            self._log.debug("incorrect frame encoding.")
            self._data.clear()
            return

        if (ver := bytes.fromhex(self._data[1:3].decode())) != BMS._RSP_VER.to_bytes():
            self._log.debug("unknown response frame version: 0x%X", int.from_bytes(ver))
            self._data.clear()
            return

        if (crc := lrc_modbus(self._data[1:-5])) != int(self._data[-5:-1], 16):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X", crc, int(self._data[-5:-1], 16)
            )
            self._data.clear()
            return

        self._data = bytearray(
            bytes.fromhex(self._data.strip(BMS._HEAD + BMS._TAIL).decode())
        )
        self._data_event.set()

    @staticmethod
    def lencs(length: int) -> int:
        """Calculate the length checksum."""
        return (sum((length >> (i * 4)) & 0xF for i in range(3)) ^ 0xF) + 1 & 0xF

    @staticmethod
    @cache
    def _cmd(cmd: int, dev_id: int = 1, data: bytes = b"") -> bytes:
        """Assemble a Seplos VB series command."""
        assert len(data) <= 0xFFF
        cdat: Final[bytes] = data + int.to_bytes(dev_id)
        frame = bytearray([BMS._CMD_VER, dev_id, 0x46, cmd])
        frame.extend(
            int.to_bytes(len(cdat) * 2 + (BMS.lencs(len(cdat) * 2) << 12), 2, "big")
        )
        frame.extend(cdat)
        frame.extend(
            int.to_bytes(lrc_modbus(bytearray(frame.hex().upper().encode())), 2, "big")
        )
        return BMS._HEAD + frame.hex().upper().encode() + BMS._TAIL

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""

        await self._await_reply(BMS._cmd(0x42))
        result: BMSSample = {"cell_count": self._data[BMS._CELL_POS]}
        temp_pos: Final[int] = BMS._CELL_POS + result.get("cell_count", 0) * 2 + 1
        result["temp_sensors"] = self._data[temp_pos]
        result["cell_voltages"] = BMS._cell_voltages(
            self._data, cells=result.get("cell_count", 0), start=BMS._CELL_POS + 1
        )
        result["temp_values"] = BMS._temp_values(
            self._data,
            values=result.get("temp_sensors", 0),
            start=temp_pos + 1,
            divider=10,
        )

        result |= BMS._decode_data(
            BMS._FIELDS, self._data, offset=temp_pos + 2 * result["temp_sensors"] + 1
        )

        await self._await_reply(BMS._cmd(0x81, 1, b"\x01\x00"), max_size=20)
        result["design_capacity"] = (
            int.from_bytes(self._data[6:8], byteorder="big", signed=False) // 10
        )

        return result



"""Module to support Dummy BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

import asyncio
from string import hexdigits
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, barr2str, crc_sum


class BMS(BaseBMS):
    """Dummy BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Wattstunde",
        "default_model": "Nova Core",
    }
    _HEAD: Final[bytes] = b"\x3a"  # beginning of frame
    _TAIL: Final[bytes] = b"\x7e"  # end of frame
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp(
            "current",
            2,
            4,
            False,
            lambda x: (x & 0x7FFF) / 1000 * (-1 if x >> 15 else 1),
        ),
        BMSDp("voltage", 10, 2, False, lambda x: x / 1000),
        BMSDp("cycles", 15, 2, False),
        BMSDp("battery_level", 17, 1, False),
        BMSDp("design_capacity", 18, 4, False, lambda x: x // 1000),
        BMSDp("cycle_charge", 22, 4, False, lambda x: x / 1000),
        BMSDp("heater", 107, 4, False, lambda x: bool(x)),
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._data_final: bytearray = bytearray()

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "manufacturer_id": 28256,
                "manufacturer_data_start": [0x41],
                "connectable": True,
            }
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("FFF0")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "FFF6"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "FFF6"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        await asyncio.wait_for(self._wait_event(), timeout=BMS.TIMEOUT)
        return BMSInfo(serial_number=barr2str(self._data_final[91:107]))

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {"power", "delta_voltage", "temperature", "runtime", "battery_charging"}
        )

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        await self._await_reply(
            BMS._HEAD + b"\x30\x31\x35\x31\x35\x30\x30\x30\x30\x45\x46\x45" + BMS._TAIL,
            wait_for_notify=False,
        )
        await super()._init_connection(char_notify)

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""

        if data.startswith(BMS._HEAD):
            self._data.clear()

        self._data += data

        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._data else "cnt.", data
        )

        if not (self._data.startswith(BMS._HEAD) and self._data.endswith(BMS._TAIL)):
            return

        if len(self._data) % 2:
            self._log.debug("incorrect frame length (%i)", len(self._data))
            return

        if not all(chr(c) in hexdigits for c in self._data[1:-1]):
            self._log.debug("incorrect frame encoding.")
            self._data.clear()
            return

        decoded = bytearray(
            b ^ 0x20 for b in bytes.fromhex(self._data[1:-1].decode("ascii"))
        )
        self._log.debug("decoded data: %s", decoded)
        # if (crc := crc_sum(decoded[:-1])) != decoded[-1]:
        #     self._log.debug("invalid checksum 0x%X != 0x%X", decoded[-1], crc)
        #     return

        self._data_final = decoded.copy()
        self._data_event.set()

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await asyncio.wait_for(self._wait_event(), timeout=BMS.TIMEOUT)

        result: BMSSample = BMS._decode_data(BMS._FIELDS, self._data_final, offset=52)
        result["cell_voltages"] = BMS._cell_voltages(
            self._data_final, cells=16, start=12
        )
        result["temp_values"] = BMS._temp_values(
            self._data_final, values=4, start=48, size=1, signed=False, offset=40
        )
        return result



"""Module to support Redodo BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, crc_sum


class BMS(BaseBMS):
    """Redodo BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Redodo",
        "default_model": "Bluetooth battery",
    }
    _HEAD_LEN: Final[int] = 3
    _MAX_CELLS: Final[int] = 16
    _MAX_TEMP: Final[int] = 3
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 12, 2, False, lambda x: x / 1000),
        BMSDp("current", 48, 4, True, lambda x: x / 1000),
        BMSDp("battery_level", 90, 2, False),
        BMSDp("cycle_charge", 62, 2, False, lambda x: x / 100),
        BMSDp("design_capacity", 64, 4, False, lambda x: x // 100),
        BMSDp("cycles", 96, 4, False),
        BMSDp("balancer", 84, 4, False),
        BMSDp("heater", 68, 4, False, bool),
        BMSDp("problem_code", 76, 4, False),
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {  # patterns required to exclude "BT-ROCC2440"
                "local_name": pattern,
                "service_uuid": BMS.uuid_services()[0],
                "manufacturer_id": 0x585A,
                "connectable": True,
            }
            for pattern in (
                "R-12*",
                "R-24*",
                "RO-12*",
                "RO-24*",
                "P-12*",
                "P-24*",
                "PQ-12*",
                "PQ-24*",
                "L-12*",  # vv *** LiTime *** vv
                "L-24*",
                "L-51*",
                "LT-12???BG-A0[7-9]*",  # LiTime based on ser#
                "LT-24???B-A00[3-9]*",
                "LT-24???B-A0[1-9]*",
                "LT-24???B-A[1-9]*",
                "LT-51*",
            )
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("ffe0")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ffe1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ffe2"

    # async def _fetch_device_info(self) -> BMSInfo: use default

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {
                "battery_charging",
                "delta_voltage",
                "cycle_capacity",
                "power",
                "runtime",
                "temperature",
            }
        )  # calculate further values from BMS provided set ones

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)

        if len(data) < 3 or not data.startswith(b"\x00\x00"):
            self._log.debug("incorrect SOF.")
            return

        if len(data) != data[2] + BMS._HEAD_LEN + 1:  # add header length and CRC
            self._log.debug("incorrect frame length (%i)", len(data))
            return

        if (crc := crc_sum(data[:-1])) != data[-1]:
            self._log.debug("invalid checksum 0x%X != 0x%X", data[len(data) - 1], crc)
            return

        self._data = data
        self._data_event.set()

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_reply(b"\x00\x00\x04\x01\x13\x55\xaa\x17")

        result: BMSSample = BMS._decode_data(
            BMS._FIELDS, self._data, byteorder="little"
        )
        result["cell_voltages"] = BMS._cell_voltages(
            self._data, cells=BMS._MAX_CELLS, start=16, byteorder="little"
        )
        result["temp_values"] = BMS._temp_values(
            self._data, values=BMS._MAX_TEMP, start=52, byteorder="little"
        )

        return result



"""Module to support Humsienk BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, barr2str, crc_sum


class BMS(BaseBMS):
    """Dummy BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Humsienk",
        "default_model": "BMC",
    }
    _HEAD: Final[bytes] = b"\xaa"  # beginning of frame
    _MIN_LEN: Final[int] = 5  # minimal frame len
    _CMDS: tuple[bytes, ...] = (b"\x20", b"\x21", b"\x22", b"\x23")
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 3, 2, False, lambda x: x / 1000, 0x21),
        BMSDp("current", 7, 4, True, lambda x: x / 1000, 0x21),
        BMSDp("battery_level", 11, 1, False, lambda x: x, 0x21),
        # BMSDp("battery_health",12, 1, False, lambda x: x, 0x21),
        BMSDp("chrg_mosfet", 9, 1, False, bool, 0x20),
        BMSDp("dischrg_mosfet", 7, 1, False, bool, 0x20),
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._data_final: dict[int, bytearray] = {}
        self._valid_reply: int = 0x00

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "local_name": "HS*",
                "service_uuid": BMS.uuid_services()[0],
                "connectable": True,
            }
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("0001")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "0003"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "0002"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        for cmd in (b"\x11", b"\xf5"):
            await self._await_reply(BMS._cmd(cmd))
        return {
            "model": barr2str(bytearray(self._data_final[0x11][3:-2])),
            "hw_version": barr2str(bytearray(self._data_final[0xF5][3:-2])),
        }

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {"power", "battery_charging", "delta_voltage"}
        )  # calculate further values from BMS provided set ones

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        """Initialize RX/TX characteristics and protocol state."""
        await super()._init_connection(char_notify)
        await self._await_reply(BMS._cmd(b"\x00"))  # init

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)

        if not data.startswith(BMS._HEAD) or ((length := len(data)) < BMS._MIN_LEN):
            self._log.debug("incorrect SOF")
            return

        if length != data[2] + BMS._MIN_LEN:
            self._log.debug("invalid frame length %d != %d", length, data[2])
            return

        if data[1] != self._valid_reply:
            self._log.debug("unexpected response (type 0x%X)", data[1])
            return

        if (crc := crc_sum(data[1:-2], 2)) != int.from_bytes(
            data[-2:], byteorder="little"
        ):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                int.from_bytes(data[-2:], byteorder="little"),
                crc,
            )
            return

        self._data_final[data[1]] = data.copy()
        self._data_event.set()

    @staticmethod
    @cache
    def _cmd(cmd: bytes) -> bytes:
        """Assemble a Seplos BMS command."""
        assert len(cmd) == 1
        frame: Final[bytes] = cmd[:1] + b"\x00"
        # head, cmd, length, 16-bit sum
        return (
            BMS._HEAD
            + frame
            + crc_sum(bytearray(frame)).to_bytes(2, byteorder="little")
        )

    async def _await_reply(
        self,
        data: bytes,
        char: int | str | None = None,
        wait_for_notify: bool = True,
        max_size: int = 0,
    ) -> None:
        """Send data to the BMS and wait for valid reply notification."""

        self._valid_reply = data[1]  # expected reply type
        await super()._await_reply(data, char, wait_for_notify, max_size)

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        for cmd in BMS._CMDS:
            await self._await_reply(BMS._cmd(cmd))

        result: BMSSample = BMS._decode_data(
            BMS._FIELDS, data=self._data_final, byteorder="little"
        )
        result["cell_voltages"] = BMS._cell_voltages(
            self._data_final[0x22], cells=24, start=3, byteorder="little"
        )
        result["temp_values"] = BMS._temp_values(
            self._data_final[0x21], values=6, start=23, size=1, byteorder="little"
        )

        return result



"""Test advertisements for aiobmsble package."""

from functools import cache
from importlib import resources
import json
from string import hexdigits
from typing import Any

from bleak.backends.scanner import AdvertisementData

from tests.bluetooth import generate_advertisement_data

type BmsAdvList = list[tuple[AdvertisementData, str, str, list[str]]]


def _json_dict_to_advdata(json_dict: dict[str, Any]) -> AdvertisementData:
    """Generate an AdvertisementData instance from a JSON dictionary."""

    if "manufacturer_data" in json_dict:
        json_dict["manufacturer_data"] = {
            int(k): bytes.fromhex(v) for k, v in json_dict["manufacturer_data"].items()
        }
    if "service_data" in json_dict:
        json_dict["service_data"] = {
            k: bytes.fromhex(v) for k, v in json_dict["service_data"].items()
        }
    if "platform_data" in json_dict:
        pdata = json_dict.get("platform_data", [])
        assert isinstance(
            pdata[0], str
        ), "first entry of platform_data needs to be string."
        parts = pdata[0].split(":")
        assert len(parts) == 6 and all(
            len(part) == 2 and all(c in hexdigits for c in part) for part in parts
        ), "first entry of platform_data only accepts MAC addresses"
        json_dict["platform_data"] = tuple(pdata)

    return generate_advertisement_data(**json_dict)


@cache
def bms_advertisements(bms_filter: str | None = None) -> BmsAdvList:
    """Provide all available BMS advertisements from test data directory.

    Load *_bms.json files from the packaged test data directory.
    If no type is given, all available advertisemetns (any type) will be returned.

    Args:
        bms_filter (str|None): BMS type for which the advertisements should be returned.

    Returns:
        BmsAdvList: List of tuples containing advertisement, mac_addr, bms type,
        and a list of comments, i.e. list[tuple[AdvertisementData, str, list[str]]]

    """
    all_data: BmsAdvList = []

    for resource in resources.files(__package__).iterdir():
        if bms_filter and not resource.name.startswith(bms_filter):
            continue
        if not resource.name.endswith("_bms.json"):
            continue
        with resource.open("r", encoding="UTF-8") as f:
            raw_data: Any = json.load(f)
            assert isinstance(raw_data, list)

            for entry in raw_data:
                assert isinstance(entry, dict)
                assert {"advertisement", "type"}.issubset(set(entry.keys()))
                adv: AdvertisementData = _json_dict_to_advdata(entry["advertisement"])
                mac_addr: str = (
                    adv.platform_data[0]
                    if isinstance(adv.platform_data, tuple)
                    and isinstance(adv.platform_data[0], str)
                    else ""
                )
                bms_type: str = entry["type"]
                comments: list[str] = entry["_comments"]

                assert isinstance(adv, AdvertisementData)
                assert isinstance(bms_type, str)
                assert isinstance(comments, list)
                assert all(isinstance(c, str) for c in comments)
                assert resource.name == f"{bms_type}.json"

                all_data.append((adv, mac_addr, bms_type, comments))
    return all_data


def ignore_advertisements() -> BmsAdvList:
    """Provide a list of advertisements that shall not be identified as a valid BMS.

    Load ignore.json files from the packaged test data directory.

    Returns:
        BmsAdvList: List of tuples containing advertisement, reason why not to detect,
        and a list of comments, i.e. list[tuple[AdvertisementData, str, list[str]]]

    """
    data: BmsAdvList = []

    with (
        resources.files(__package__)
        .joinpath("ignore.json")
        .open("r", encoding="UTF-8") as f
    ):
        raw_data: Any = json.load(f)
        assert isinstance(raw_data, list)

        for entry in raw_data:
            assert isinstance(entry, dict)
            assert {"advertisement", "reason"}.issubset(set(entry.keys()))
            adv: AdvertisementData = _json_dict_to_advdata(entry["advertisement"])
            mac_addr: str = adv.platform_data[0] if "platform" in adv else ""
            reason: str = entry["reason"]
            comments: list[str] = entry["_comments"]

            assert isinstance(adv, AdvertisementData)
            assert isinstance(reason, str)
            assert isinstance(comments, list)
            assert all(isinstance(c, str) for c in comments)

            data.append((adv, mac_addr, reason, comments))

    return data



"""Module to support Braun Power BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, barr2str


class BMS(BaseBMS):
    """Braun Power BMS class implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Braun Power",
        "default_model": "smart BMS",
    }
    _HEAD: Final[bytes] = b"\x7b"  # header for responses
    _TAIL: Final[int] = 0x7D  # tail for command
    _MIN_LEN: Final[int] = 4  # minimum frame size
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("cell_count", 3, 1, False, lambda x: x, 0x2),
        BMSDp("temp_sensors", 3, 1, False, lambda x: x, 0x3),
        BMSDp("voltage", 5, 2, False, lambda x: x / 100, 0x1),
        BMSDp("current", 13, 2, True, lambda x: x / 100, 0x1),
        BMSDp("battery_level", 4, 1, False, lambda x: x, 0x1),
        BMSDp("cycle_charge", 15, 2, False, lambda x: x / 100, 0x1),
        BMSDp("design_capacity", 17, 2, False, lambda x: x // 100, 0x1),
        BMSDp("cycles", 23, 2, False, lambda x: x, 0x1),
        BMSDp("problem_code", 31, 2, False, lambda x: x, 0x1),
        BMSDp("balancer", 20, 2, False, lambda x: x, 0x1),
    )
    _CMDS: Final[set[int]] = {field.idx for field in _FIELDS}
    _INIT_CMDS: Final[set[int]] = {
        0x74,  # SW version
        0xF4,  # BMS program version
        0xF5,  # BMS boot version
    }

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Intialize private BMS members."""
        super().__init__(ble_device, keep_alive)
        self._data_final: dict[int, bytearray] = {}
        self._exp_reply: tuple[int] = (0x01,)

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            MatcherPattern(
                local_name=pattern,
                service_uuid=BMS.uuid_services()[0],
                manufacturer_id=0x7B,
                connectable=True,
            )
            for pattern in ("HSKS-*", "BL-*")
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("ff00")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ff01"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ff02"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        for cmd in BMS._INIT_CMDS:
            self._exp_reply = (cmd,)
            await self._await_reply(BMS._cmd(cmd))
        return {
            "sw_version": ".".join(str(x) for x in self._data_final[0xF4][3:6]),
            "hw_version": barr2str(self._data_final[0x74][3:-1]),
        }

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {
                "power",
                "battery_charging",
                "cycle_capacity",
                "runtime",
                "delta_voltage",
                "temperature",
            }
        )

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        # check if answer is a heading of valid response type
        if (
            data.startswith(BMS._HEAD)
            and len(self._data) >= BMS._MIN_LEN
            and data[1] in {*BMS._CMDS, *BMS._INIT_CMDS}
            and len(self._data) >= BMS._MIN_LEN + self._data[2]
        ):
            self._data = bytearray()

        self._data += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._data else "cnt.", data
        )

        # verify that data is long enough
        if (
            len(self._data) < BMS._MIN_LEN
            or len(self._data) < BMS._MIN_LEN + self._data[2]
        ):
            return

        # check correct frame ending
        if self._data[-1] != BMS._TAIL:
            self._log.debug("incorrect frame end (length: %i).", len(self._data))
            self._data.clear()
            return

        if self._data[1] not in self._exp_reply:
            self._log.debug("unexpected command 0x%02X", self._data[1])
            self._data.clear()
            return

        # check if response length matches expected length
        if len(self._data) != BMS._MIN_LEN + self._data[2]:
            self._log.debug("wrong data length (%i): %s", len(self._data), self._data)
            self._data.clear()
            return

        self._data_final[self._data[1]] = self._data
        self._data_event.set()

    @staticmethod
    def _cmd(cmd: int, data: bytes = b"") -> bytes:
        """Assemble a Braun Power BMS command."""
        assert len(data) <= 255, "data length must be a single byte."
        return bytes([*BMS._HEAD, cmd, len(data), *data, BMS._TAIL])

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        """Connect to the BMS and setup notification if not connected."""
        await super()._init_connection()
        await self._fetch_device_info()

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        self._data_final.clear()
        for cmd in BMS._CMDS:
            self._exp_reply = (cmd,)
            await self._await_reply(BMS._cmd(cmd))

        data: BMSSample = BMS._decode_data(BMS._FIELDS, self._data_final)
        data["cell_voltages"] = BMS._cell_voltages(
            self._data_final[0x2], cells=data.get("cell_count", 0), start=4
        )
        data["temp_values"] = BMS._temp_values(
            self._data_final[0x3],
            values=data.get("temp_sensors", 0),
            start=4,
            offset=2731,
            divider=10,
        )

        return data



"""Module to support Renogy Pro BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, MatcherPattern
from aiobmsble.bms.renogy_bms import BMS as RenogyBMS


class BMS(RenogyBMS):
    """Renogy Pro battery class implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Renogy",
        "default_model": "BT battery pro",
    }
    _HEAD: bytes = b"\xff\x03"  # SOP, read fct (x03)
    FIELDS: tuple[BMSDp, ...] = (
        BMSDp("voltage", 5, 2, False, lambda x: x / 10),
        BMSDp("current", 3, 2, True, lambda x: x / 10),
        BMSDp("design_capacity", 11, 4, False, lambda x: x // 1000),
        BMSDp("cycle_charge", 7, 4, False, lambda x: x / 1000),
        BMSDp("cycles", 15, 2, False, lambda x: x),
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Intialize private BMS members."""
        super().__init__(ble_device, keep_alive)
        self._char_write_handle: int = -1

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "local_name": "RNGRBP*",
                "manufacturer_id": 0xE14C,
                "connectable": True,
            },
        ]

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        """Initialize RX/TX characteristics and protocol state."""
        char_notify_handle: int = -1
        self._char_write_handle = -1
        assert char_notify is None, "char_notify not used for Renogy Pro BMS"

        for service in self._client.services:
            self._log.debug(
                "service %s (#%i): %s",
                service.uuid,
                service.handle,
                service.description,
            )
            for char in service.characteristics:
                self._log.debug(
                    "characteristic %s (#%i): %s",
                    char.uuid,
                    char.handle,
                    char.properties,
                )
                if (
                    service.uuid == BMS.uuid_services()[0]
                    and char.uuid == normalize_uuid_str(BMS.uuid_tx())
                    and any(
                        prop in char.properties
                        for prop in ("write", "write-without-response")
                    )
                ):
                    self._char_write_handle = char.handle
                if (
                    service.uuid == BMS.uuid_services()[1]
                    and char.uuid == normalize_uuid_str(BMS.uuid_rx())
                    and "notify" in char.properties
                ):
                    char_notify_handle = char.handle

        if char_notify_handle == -1 or self._char_write_handle == -1:
            self._log.debug("failed to detect characteristics.")
            await self._client.disconnect()
            raise ConnectionError(
                f"Failed to detect characteristics from {self.name}."
            )
        self._log.debug(
            "using characteristics handle #%i (notify), #%i (write).",
            char_notify_handle,
            self._char_write_handle,
        )

        await super()._init_connection(char_notify_handle)

    async def _await_reply(
        self,
        data: bytes,
        char: int | str | None = None,
        wait_for_notify: bool = True,
        max_size: int = 0,
    ) -> None:
        """Send data to the BMS and wait for valid reply notification."""

        await super()._await_reply(
            data, self._char_write_handle, wait_for_notify, max_size
        )



"""Test the aiobmsble test data functions."""

from bleak.backends.scanner import AdvertisementData

from aiobmsble.test_data import (
    _json_dict_to_advdata,
    bms_advertisements,
    ignore_advertisements,
)


def test_test_data_availability() -> None:
    """That fixtures work, safety measure to avoid package corruption."""
    assert len(bms_advertisements()) > 0
    assert len(ignore_advertisements()) > 0


def test_json_dict_to_advdata() -> None:
    """Check that conversion delivers correct types."""

    data = {
        "local_name": "dummy_bms",
        "rssi": -73,
        "manufacturer_data": {"12345": "00112233445566778899AABBCCDDEEFF"},
        "service_data": {
            "0000262a-0000-1000-8000-00805f9b34fb": "112233445566778899AABB"
        },
        "service_uuids": ["6e400001-b5a3-f393-e0a9-e50e24dcca9e"],
        "tx_power": 21,
        "platform_data": ("00:11:22:aa:bb:cc", 123, "test", {"check": "ok"}),
    }
    assert _json_dict_to_advdata(data) == AdvertisementData(
        local_name="dummy_bms",
        rssi=-73,
        manufacturer_data={
            12345: b"\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff"
        },
        service_data={
            "0000262a-0000-1000-8000-00805f9b34fb": b"\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb"
        },
        service_uuids=["6e400001-b5a3-f393-e0a9-e50e24dcca9e"],
        tx_power=21,
        platform_data=("00:11:22:aa:bb:cc", 123, "test", {"check": "ok"}),
    )



"""Module to support TDT BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, crc_modbus


class BMS(BaseBMS):
    """TDT BMS implementation."""

    INFO: BMSInfo = {"default_manufacturer": "TDT", "default_model": "smart BMS"}
    _UUID_CFG: Final[str] = "fffa"
    _HEAD: Final[int] = 0x7E
    _CMD_HEADS: list[int] = [0x7E, 0x1E]  # alternative command head
    _TAIL: Final[int] = 0x0D
    _CMD_VER: Final[int] = 0x00
    _RSP_VER: Final[frozenset[int]] = frozenset({0x00, 0x04})
    _CELL_POS: Final[int] = 0x8
    _INFO_LEN: Final[int] = 10  # minimal frame length
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 2, 2, False, lambda x: x / 100, 0x8C),
        BMSDp(
            "current",
            0,
            2,
            False,
            lambda x: (x & 0x3FFF) / 10 * (-1 if x >> 15 else 1),
            0x8C,
        ),
        BMSDp("cycle_charge", 4, 2, False, lambda x: x / 10, 0x8C),
        BMSDp("battery_level", 13, 1, False, lambda x: x, 0x8C),
        BMSDp("cycles", 8, 2, False, lambda x: x, 0x8C),
    )  # problem code, switches are not included in the list, but extra
    _CMDS: Final[list[int]] = [*list({field.idx for field in _FIELDS}), 0x8D]

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._data_final: dict[int, bytearray] = {}
        self._cmd_heads: list[int] = BMS._CMD_HEADS
        self._exp_len: int = 0

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [{"manufacturer_id": 54976, "connectable": True}]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("fff0")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "fff1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "fff2"

    # async def _fetch_device_info(self) -> BMSInfo: use default

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {
                "battery_charging",
                "cycle_capacity",
                "delta_voltage",
                "power",
                "runtime",
                "temperature",
            }
        )  # calculate further values from BMS provided set ones

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        await self._await_reply(
            data=b"HiLink", char=BMS._UUID_CFG, wait_for_notify=False
        )
        if (
            ret := int.from_bytes(await self._client.read_gatt_char(BMS._UUID_CFG))
        ) != 0x1:
            self._log.debug("error unlocking BMS: %X", ret)

        await super()._init_connection()

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)

        if (
            len(data) > BMS._INFO_LEN
            and data[0] == BMS._HEAD
            and len(self._data) >= self._exp_len
        ):
            self._exp_len = BMS._INFO_LEN + int.from_bytes(data[6:8])
            self._data = bytearray()

        self._data += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._data else "cnt.", data
        )

        # verify that data is long enough
        if len(self._data) < max(BMS._INFO_LEN, self._exp_len):
            return

        if self._data[-1] != BMS._TAIL:
            self._log.debug("frame end incorrect: %s", self._data)
            return

        if self._data[1] not in BMS._RSP_VER:
            self._log.debug("unknown frame version: V%.1f", self._data[1] / 10)
            return

        if self._data[4]:
            self._log.debug("BMS reported error code: 0x%X", self._data[4])
            return

        if (crc := crc_modbus(self._data[:-3])) != int.from_bytes(
            self._data[-3:-1], "big"
        ):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                int.from_bytes(self._data[-3:-1], "big"),
                crc,
            )
            return
        self._data_final[self._data[5]] = self._data
        self._data_event.set()

    @staticmethod
    @cache
    def _cmd(cmd: int, data: bytearray = bytearray(), cmd_head: int = _HEAD) -> bytes:
        """Assemble a TDT BMS command."""
        assert cmd in (0x8C, 0x8D, 0x92)  # allow only read commands

        frame = bytearray([cmd_head, BMS._CMD_VER, 0x1, 0x3, 0x0, cmd])
        frame += len(data).to_bytes(2, "big", signed=False) + data
        frame += crc_modbus(frame).to_bytes(2, "big") + bytes([BMS._TAIL])

        return bytes(frame)

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""

        for head in self._cmd_heads:
            try:
                for cmd in BMS._CMDS:
                    await self._await_reply(BMS._cmd(cmd, cmd_head=head))
                self._cmd_heads = [head]  # set to single head for further commands
                break
            except TimeoutError:
                ...  # try next command head
        else:
            raise TimeoutError

        result: BMSSample = {"cell_count": self._data_final[0x8C][BMS._CELL_POS]}
        result["temp_sensors"] = self._data_final[0x8C][
            BMS._CELL_POS + result["cell_count"] * 2 + 1
        ]

        result["cell_voltages"] = BMS._cell_voltages(
            self._data_final[0x8C],
            cells=result.get("cell_count", 0),
            start=BMS._CELL_POS + 1,
        )
        result["temp_values"] = BMS._temp_values(
            self._data_final[0x8C],
            values=result["temp_sensors"],
            start=BMS._CELL_POS + result.get("cell_count", 0) * 2 + 2,
            signed=False,
            offset=2731,
            divider=10,
        )
        idx: Final[int] = result.get("cell_count", 0) + result.get("temp_sensors", 0)

        result |= BMS._decode_data(
            BMS._FIELDS, self._data_final, offset=BMS._CELL_POS + idx * 2 + 2
        )
        result["problem_code"] = int.from_bytes(
            self._data_final[0x8D][BMS._CELL_POS + idx + 6 : BMS._CELL_POS + idx + 8]
        )
        mosfets: Final[int] = self._data_final[0x8D][BMS._CELL_POS + idx + 8]
        result |= {
            "chrg_mosfet": bool(mosfets & 0x4),
            "dischrg_mosfet": bool(mosfets & 0x2),
        }

        self._data_final.clear()

        return result



"""Module to support ECO-WORTHY BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

import asyncio
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, crc_modbus


class BMS(BaseBMS):
    """ECO-WORTHY BMS implementation."""

    INFO: BMSInfo = {"default_manufacturer": "ECO-WORTHY", "default_model": "BW02"}
    _HEAD: Final[tuple[bytes, ...]] = (b"\xa1", b"\xa2")
    _CELL_POS: Final[int] = 14
    _TEMP_POS: Final[int] = 80
    _FIELDS_V1: Final[tuple[BMSDp, ...]] = (
        BMSDp("battery_level", 16, 2, False, lambda x: x, 0xA1),
        BMSDp("voltage", 20, 2, False, lambda x: x / 100, 0xA1),
        BMSDp("current", 22, 2, True, lambda x: x / 100, 0xA1),
        BMSDp("problem_code", 51, 2, False, lambda x: x, 0xA1),
        BMSDp("design_capacity", 26, 2, False, lambda x: x // 100, 0xA1),
        BMSDp("cell_count", _CELL_POS, 2, False, lambda x: x, 0xA2),
        BMSDp("temp_sensors", _TEMP_POS, 2, False, lambda x: x, 0xA2),
        # ("cycles", 0xA1, 8, 2, False, lambda x: x),
    )
    _FIELDS_V2: Final[tuple[BMSDp, ...]] = tuple(
        BMSDp(
            *field[:-2],
            (lambda x: x / 10) if field.key == "current" else field.fct,
            field.idx,
        )
        for field in _FIELDS_V1
    )

    _CMDS: Final[set[int]] = set({field.idx for field in _FIELDS_V1})

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._mac_head: Final[tuple[bytes, ...]] = tuple(
            int(self._ble_device.address.replace(":", ""), 16).to_bytes(6) + head
            for head in BMS._HEAD
        )
        self._data_final: dict[int, bytearray] = {}

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [MatcherPattern(local_name="ECO-WORTHY 02_*", connectable=True)] + [
            MatcherPattern(
                local_name=pattern,
                service_uuid=BMS.uuid_services()[0],
                connectable=True,
            )
            for pattern in ("DCHOUSE*", "ECO-WORTHY*")
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("fff0")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "fff1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        raise NotImplementedError

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {
                "battery_charging",
                "cycle_charge",
                "cycle_capacity",
                "delta_voltage",
                "power",
                "runtime",
                "temperature",
            }
        )  # calculate further values from BMS provided set ones

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)

        if not data.startswith(BMS._HEAD + self._mac_head):
            self._log.debug("invalid frame type: '%s'", data[0:1].hex())
            return

        if (crc := crc_modbus(data[:-2])) != int.from_bytes(data[-2:], "little"):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                int.from_bytes(data[-2:], "little"),
                crc,
            )
            return

        # copy final data without message type and adapt to protocol type
        shift: Final[bool] = data.startswith(self._mac_head)
        self._data_final[data[6 if shift else 0]] = bytearray(
            2 if shift else 0
        ) + bytes(data)
        if BMS._CMDS.issubset(self._data_final.keys()):
            self._data_event.set()

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""

        self._data_final.clear()
        self._data_event.clear()  # clear event to ensure new data is acquired
        await asyncio.wait_for(self._wait_event(), timeout=BMS.TIMEOUT)

        result: BMSSample = BMS._decode_data(
            (
                BMS._FIELDS_V1
                if self._data_final[0xA1].startswith(BMS._HEAD)
                else BMS._FIELDS_V2
            ),
            self._data_final,
        )

        result["cell_voltages"] = BMS._cell_voltages(
            self._data_final[0xA2],
            cells=result.get("cell_count", 0),
            start=BMS._CELL_POS + 2,
        )
        result["temp_values"] = BMS._temp_values(
            self._data_final[0xA2],
            values=result.get("temp_sensors", 0),
            start=BMS._TEMP_POS + 2,
            divider=10,
        )

        return result



"""Module to support Vatrer BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, crc_modbus


class BMS(BaseBMS):
    """Vatrer BMS implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Vatrer", "default_model": "smart BMS"}
    _HEAD: Final[bytes] = b"\x02\x03"  # beginning of frame
    _FRAME_LEN: Final[int] = 5  # head + len + CRC
    _MAX_CELLS: Final[int] = 0x1F
    _MAX_TEMP: Final[int] = 6
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 3, 2, False, lambda x: x / 100, 0x28),
        BMSDp("current", 5, 4, True, lambda x: x / 100, 0x28),
        BMSDp("battery_level", 9, 2, False, idx=0x28),
        BMSDp("cycle_charge", 11, 2, False, lambda x: x / 100, 0x28),
        BMSDp("cell_count", 3, 2, False, lambda x: min(x, BMS._MAX_CELLS), 0x3E),
        BMSDp("temp_sensors", 3, 2, False, lambda x: min(x, BMS._MAX_TEMP), 0x24),
        BMSDp("cycles", 15, 2, False, lambda x: x, 0x28),
        BMSDp("delta_voltage", 29, 2, False, lambda x: x / 1000, 0x28),
        BMSDp("problem", 17, 15, False, lambda x: (x != 0), 0x24),
        BMSDp("chrg_mosfet", 32, 1, False, lambda x: bool(x & 0x10), 0x24),
        BMSDp("dischrg_mosfet", 32, 1, False, lambda x: bool(x & 0x20), 0x24),
        BMSDp("balancer", 35, 4, False, lambda x: x, 0x24),
    )
    _RESPS: Final[set[int]] = {field.idx for field in _FIELDS}
    _CMDS: Final[tuple[tuple[int, int], ...]] = (
        (0x0, 0x14),
        (0x34, 0x12),
        (0x15, 0x1F),
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._data_final: dict[int, bytearray] = {}

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {  # name is likely YYMMDDVVVAAAAxx (date, V, Ah)
                "local_name": "[2-9]???[0-3]?512??00??",
                "service_uuid": BMS.uuid_services()[0],
                "connectable": True,
            }
        ]

    # async def _fetch_device_info(self) -> BMSInfo: use default

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return ["6e400001-b5a3-f393-e0a9-e50e24dcca9e"]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "6e400003-b5a3-f393-e0a9-e50e24dcca9e"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "6e400002-b5a3-f393-e0a9-e50e24dcca9e"

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {"power", "battery_charging", "temperature", "runtime", "cycle_capacity"}
        )  # calculate further values from BMS provided set ones

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)

        if not data.startswith(BMS._HEAD):
            self._log.debug("incorrect SOF")
            return

        if len(data) < BMS._FRAME_LEN or len(data) != data[2] + BMS._FRAME_LEN:
            self._log.debug("incorrect frame length")
            return

        if (crc := crc_modbus(data[:-2])) != int.from_bytes(
            data[-2:], byteorder="little"
        ):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                int.from_bytes(data[-2:], "little"),
                crc,
            )
            return

        self._data_final[data[2]] = data.copy()
        self._data_event.set()

    @staticmethod
    @cache
    def _cmd(addr: int, words: int) -> bytes:
        """Assemble a Vatrer BMS command."""
        frame: bytearray = (
            bytearray(BMS._HEAD)
            + addr.to_bytes(2, byteorder="big")
            + words.to_bytes(2, byteorder="big")
        )
        frame.extend(crc_modbus(frame).to_bytes(2, "little"))
        return bytes(frame)

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        for addr, length in BMS._CMDS:
            await self._await_reply(BMS._cmd(addr, length))
        if not BMS._RESPS.issubset(set(self._data_final.keys())):
            self._log.debug("Incomplete data set %s", self._data_final.keys())
            raise TimeoutError("BMS data incomplete.")

        result: BMSSample = BMS._decode_data(BMS._FIELDS, self._data_final)
        result["cell_voltages"] = BMS._cell_voltages(
            self._data_final[0x3E], cells=result.get("cell_count", 0), start=5
        )
        result["temp_values"] = BMS._temp_values(
            self._data_final[0x24], values=result.get("temp_sensors", 0) + 2, start=5
        )  # MOS sensor is last (pos 6 of 4)

        return result



"""Test clases for package exaples."""

from collections.abc import Callable
from typing import NoReturn

from bleak.backends.device import BLEDevice
from bleak.exc import BleakError
import pytest

import examples.minimal
from tests.bluetooth import generate_ble_device


@pytest.fixture(name="patch_find_device_by_name")
def fixture_find_device_by_name(monkeypatch: pytest.MonkeyPatch):
    """Fixture to patch BleakClient with a given MockClient."""

    def _patch() -> None:
        monkeypatch.setattr(
            "examples.minimal.BleakScanner.find_device_by_name",
            TestMinimal.mock_find_device_by_name,
        )

    return _patch


class TestMinimal:
    """Test the minimal.py example."""

    @staticmethod
    async def mock_find_device_by_name(
        name: str, timeout: float = 10.0, **_kwargs
    ) -> BLEDevice | None:
        """Mock function that returns a BleakDevice for 'Dummy device'."""
        assert timeout > 0
        if name == "Dummy device":
            return generate_ble_device("11:22:33:44:55:66", name)
        return None

    @staticmethod
    async def test_main(
        patch_find_device_by_name: Callable[[], None],
        patch_bleak_client: Callable,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """Test that minimal example returns BMS data."""
        patch_find_device_by_name()
        patch_bleak_client()
        await examples.minimal.main("Dummy device")
        assert "Found device: Dummy device (11:22:33:44:55:66)" in caplog.text
        assert (
            "BMS data: {'voltage': 12,\n\t'current': 1.5,\n\t'temperature': 27.182,\n"
            "\t'power': 18.0,\n\t'battery_charging': True,\n\t'problem': False}\n"
            in caplog.text
        )

    @staticmethod
    async def test_main_no_dev(
        patch_find_device_by_name: Callable[[], None],
        patch_bleak_client: Callable,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """Test that minimal example handles if the device with given name is not found."""

        patch_find_device_by_name()
        patch_bleak_client()
        await examples.minimal.main("wrong_name")
        assert "Device 'wrong_name' not found." in caplog.text

    @staticmethod
    async def test_main_bt_err(
        monkeypatch: pytest.MonkeyPatch,
        patch_find_device_by_name: Callable[[], None],
        patch_bleak_client: Callable,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """Test that minimal example handles if a Bleak error occurs."""

        def _raise_bleak_error(_self) -> NoReturn:
            raise BleakError

        patch_find_device_by_name()
        # Ensure BMS attribute exists on the examples.minimal module before monkeypatching
        assert hasattr(examples.minimal, "BMS")
        monkeypatch.setattr(examples.minimal.BMS, "async_update", _raise_bleak_error)
        patch_bleak_client()
        await examples.minimal.main("Dummy device")
        assert "Failed to update BMS: BleakError" in caplog.text



"""Module to support Pro BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

import asyncio
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS


class BMS(BaseBMS):
    """Pro BMS Smart Shunt class implementation."""

    INFO: BMSInfo = {"default_manufacturer": "Pro BMS", "default_model": "Smart Shunt"}
    _HEAD: Final[bytes] = b"\x55\xaa"
    _MIN_LEN: Final[int] = 5
    _INIT_RESP: Final[int] = 0x03
    _RT_DATA: Final[int] = 0x04

    # Commands from btsnoop capture
    _CMD_INIT: Final[bytes] = bytes.fromhex("0a0101558004077f648e682b")
    _CMD_ACK: Final[bytes] = bytes.fromhex("070101558040000095")
    _CMD_DATA_STREAM: Final[bytes] = bytes.fromhex("070101558042000097")
    # command that triggers data streaming (Function 0x43)
    _CMD_TRIGGER_DATA: Final[bytes] = bytes.fromhex("0901015580430000120084")

    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 8, 2, False, lambda x: x / 100),
        BMSDp(
            "current",
            12,
            4,
            False,
            lambda x: ((x & 0xFFFF) / 1000) * (-1 if (x >> 24) & 0x80 else 1),
        ),
        BMSDp("problem_code", 15, 4, False, lambda x: x & 0x7F),
        BMSDp(
            "temperature",
            16,
            3,
            False,
            lambda x: ((x & 0xFFFF) / 10) * (-1 if x >> 16 else 1),
        ),
        BMSDp("cycle_charge", 20, 4, False, lambda x: x / 100),
        BMSDp("battery_level", 24, 1, False, lambda x: x),
        BMSDp("power", 32, 4, False, lambda x: x / 100),
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize private BMS members."""
        super().__init__(ble_device, keep_alive)
        self._valid_reply: int = BMS._RT_DATA

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            MatcherPattern(
                local_name="Pro BMS",
                service_uuid=BMS.uuid_services()[0],
                connectable=True,
            )
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("fff0")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "fff4"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "fff3"

    # async def _fetch_device_info(self) -> BMSInfo: unknown, use default

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset({"battery_charging", "cycle_capacity", "runtime"})

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        self._log.debug("RX BLE data: %s", data)

        if len(data) < BMS._MIN_LEN or not data.startswith(BMS._HEAD):
            self._log.debug("Invalid packet header")
            return

        if data[3] != self._valid_reply:
            self._log.debug("unexpected response (type 0x%X)", data[3])
            return

        if len(data) != data[2] + BMS._MIN_LEN:
            self._log.debug("incorrect frame length: %i).", len(self._data))
            return

        self._data = data
        self._data_event.set()

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        """Initialize RX/TX characteristics and protocol state."""
        await super()._init_connection()
        self._valid_reply = BMS._INIT_RESP

        # Step 1: Send initialization command and await response
        await self._await_reply(BMS._HEAD + BMS._CMD_INIT)

        # Step 2: Send ACK command
        # Step 3: Send data stream command
        # Step 4: Send trigger data command 0x43 - start RT data stream
        for cmd in (BMS._CMD_ACK, BMS._CMD_DATA_STREAM, BMS._CMD_TRIGGER_DATA):
            await self._await_reply(BMS._HEAD + cmd, wait_for_notify=False)

        self._valid_reply = BMS._RT_DATA

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""

        self._data_event.clear()  # Clear the event to ensure fresh data on each update
        try:
            # Wait for new data packet
            await asyncio.wait_for(self._wait_event(), timeout=BMS.TIMEOUT)
        except TimeoutError:
            await self.disconnect()
            raise

        result: BMSSample = BMS._decode_data(
            BMS._FIELDS, self._data, byteorder="little"
        )
        result["power"] = result.get("power", 0) * (
            -1 if result.get("current", 0) < 0 else 1
        )
        return result



"""Module to support ABC BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

import contextlib
from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, barr2str, crc8


class BMS(BaseBMS):
    """ABC BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Chunguang Song",
        "default_model": "ABC-BMS",
    }
    _HEAD_CMD: Final[int] = 0xEE
    _HEAD_RESP: Final[bytes] = b"\xcc"
    _INFO_LEN: Final[int] = 0x14
    _EXP_REPLY: Final[dict[int, set[int]]] = {  # wait for these replies
        0xC0: {0xF1},
        0xC1: {0xF0, 0xF2},
        0xC2: {0xF0, 0xF3, 0xF4},  # 4 cells per F4 message
        0xC3: {0xF5, 0xF6, 0xF7, 0xF8, 0xFA},
        0xC4: {0xF9},
    }
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("temp_sensors", 4, 1, False, lambda x: x, 0xF2),
        BMSDp("voltage", 2, 3, False, lambda x: x / 1000, 0xF0),
        BMSDp("current", 5, 3, True, lambda x: x / 1000, 0xF0),
        BMSDp("design_capacity", 8, 3, False, lambda x: x // 1000, 0xF0),
        BMSDp("battery_level", 16, 1, False, lambda x: x, 0xF0),
        BMSDp("cycle_charge", 11, 3, False, lambda x: x / 1000, 0xF0),
        BMSDp("cycles", 14, 2, False, lambda x: x, 0xF0),
        BMSDp(  # only first bit per byte is used
            "problem_code",
            2,
            16,
            False,
            lambda x: sum(((x >> (i * 8)) & 1) << i for i in range(16)),
            0xF9,
        ),
        BMSDp("chrg_mosfet", 2, 1, False, bool, 0xF2),
        BMSDp("dischrg_mosfet", 3, 1, False, bool, 0xF2),
        BMSDp("heater", 8, 1, False, bool, 0xF3),
        BMSDp("balancer", 10, 1, False, lambda x: x, 0xF3),
    )
    _RESPS: Final[set[int]] = {field.idx for field in _FIELDS} | {0xF4}  # cell voltages

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._data_final: dict[int, bytearray] = {}
        self._exp_reply: set[int] = set()

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "local_name": pattern,
                "service_uuid": normalize_uuid_str("fff0"),
                "connectable": True,
            }
            for pattern in ("ABC-*", "SOK-*", "NB-*", "Hoover")
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("ffe0")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ffe1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ffe2"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        info: BMSInfo = await super()._fetch_device_info()
        self._exp_reply = BMS._EXP_REPLY[0xC0]
        await self._await_reply(BMS._cmd(bytes([0xC0])))
        info |= {"model": barr2str(self._data_final[0xF1][2:-1])}
        return info

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {
                "battery_charging",
                "cycle_capacity",
                "delta_voltage",
                "power",
                "runtime",
                "temperature",
            }
        )  # calculate further values from BMS provided set ones

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)

        if not data.startswith(BMS._HEAD_RESP):
            self._log.debug("Incorrect frame start")
            return

        if len(data) != BMS._INFO_LEN:
            self._log.debug("Incorrect frame length")
            return

        if (crc := crc8(data[:-1])) != data[-1]:
            self._log.debug("invalid checksum 0x%X != 0x%X", data[-1], crc)
            return

        if data[1] == 0xF4 and 0xF4 in self._data_final:
            # expand cell voltage frame with all parts
            self._data_final[0xF4] = bytearray(self._data_final[0xF4][:-2] + data[2:])
        else:
            self._data_final[data[1]] = data.copy()

        self._exp_reply.discard(data[1])

        if not self._exp_reply:  # check if all expected replies are received
            self._data_event.set()

    @staticmethod
    @cache
    def _cmd(cmd: bytes) -> bytes:
        """Assemble a ABC BMS command."""
        frame = bytearray([BMS._HEAD_CMD, cmd[0], 0x00, 0x00, 0x00])
        frame.append(crc8(frame))
        return bytes(frame)

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        self._data_final.clear()
        for cmd in (0xC1, 0xC2, 0xC4):
            self._exp_reply.update(BMS._EXP_REPLY[cmd])
            with contextlib.suppress(TimeoutError):
                await self._await_reply(BMS._cmd(bytes([cmd])))

        # check all repsonses are here, 0xF9 is not mandatory (not all BMS report it)
        self._data_final.setdefault(0xF9, bytearray())
        if not BMS._RESPS.issubset(set(self._data_final.keys())):
            self._log.debug("Incomplete data set %s", self._data_final.keys())
            raise TimeoutError("BMS data incomplete.")

        result: BMSSample = BMS._decode_data(
            BMS._FIELDS, self._data_final, byteorder="little"
        )
        return result | {
            "cell_voltages": BMS._cell_voltages(  # every second value is the cell idx
                self._data_final[0xF4],
                cells=(len(self._data_final[0xF4]) - 4) // 2,
                start=3,
                byteorder="little",
                size=2,
            )[::2],
            "temp_values": BMS._temp_values(
                self._data_final[0xF2],
                start=5,
                values=result.get("temp_sensors", 0),
                byteorder="little",
            ),
        }



"""Module to support PACEEX BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, barr2str, crc_modbus


class BMS(BaseBMS):
    """PACEEX BMS implementation."""

    # TODO: implement multi battery pack

    INFO: BMSInfo = {
        "default_manufacturer": "PeiCheng Technology",
        "default_model": "PACEEX Smart BMS",
    }
    _HEAD: Final[bytes] = b"\x9a"
    _TAIL: Final[bytes] = b"\x9d"
    _FRM_TYPE: Final[slice] = slice(1, 7)
    _MIN_LEN: Final[int] = 11  # minimal frame length
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("current", 1, 4, True, lambda x: x / 100),
        BMSDp("voltage", 5, 4, False, lambda x: x / 100),
        BMSDp("cycle_charge", 9, 4, False, lambda x: x / 100),
        BMSDp("design_capacity", 13, 4, False, lambda x: x // 100),
        BMSDp("battery_level", 21, 1, False),
        # BMSDp("battery_health", 22, 1, False, lambda x: x),
        BMSDp("pack_count", 0, 1, False),
        BMSDp("cycles", 23, 4, False),
        # BMSDp("problem_code", 1, 9, False, lambda x: x & 0xFFFF00FF00FF0000FF, EIC_LEN),
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._valid_reply: bytes = b""  # expected reply type

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [{"local_name": "PC-????", "connectable": True}]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("fff0")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "fff1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "fff2"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""
        result: BMSInfo = BMSInfo()
        await self._await_reply(self._cmd(b"\x00\x00\x00\x02\x00\x00"))
        length: int = self._data[8]
        result["serial_number"] = barr2str(self._data[9 : 9 + length])
        await self._await_reply(self._cmd(b"\x00\x00\x00\x01\x00\x00"))
        result["sw_version"] = barr2str(self._data[10 : 10 + self._data[9]])
        result["hw_version"] = barr2str(self._data[65 : 65 + self._data[64]])
        return result

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {
                "delta_voltage",
                "cycle_capacity",
                "power",
                "runtime",
                "battery_charging",
                "temperature",
            }
        )

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""
        self._log.debug("RX BLE data: %s", data)

        if not data.startswith(BMS._HEAD):
            self._log.debug("incorrect SOF")
            return

        if len(data) < BMS._MIN_LEN or len(data) != BMS._MIN_LEN + data[7]:
            self._log.debug("incorrect frame length")
            return

        if (crc := crc_modbus(data[:-3])) != int.from_bytes(
            data[-3:-1], byteorder="big"
        ):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                int.from_bytes(data[-3:-1], byteorder="big"),
                crc,
            )
            return

        if data[BMS._FRM_TYPE] != self._valid_reply:
            self._log.debug("unexpected response")
            return

        self._data = data.copy()
        self._data_event.set()

    @staticmethod
    @cache
    def _cmd(cmd: bytes, data: bytes = b"") -> bytes:
        """Assemble a Pace BMS command."""
        frame: bytearray = bytearray(BMS._HEAD) + cmd + len(data).to_bytes(1) + data
        frame += int.to_bytes(crc_modbus(frame), 2, byteorder="big") + BMS._TAIL
        return bytes(frame)

    async def _await_reply(
        self,
        data: bytes,
        char: int | str | None = None,
        wait_for_notify: bool = True,
        max_size: int = 0,
    ) -> None:
        """Send data to the BMS and wait for valid reply notification."""

        self._valid_reply = data[BMS._FRM_TYPE]  # expected reply type
        await super()._await_reply(data, char, wait_for_notify, max_size)

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_reply(BMS._cmd(b"\x00\x00\x0a\x00\x00\x00"))
        result = BMS._decode_data(BMS._FIELDS, self._data, byteorder="big", offset=8)
        await self._await_reply(BMS._cmd(b"\x00\x00\x0a\x02\x00\x00", b"\x01\x01"))
        result["cell_count"] = self._data[11]
        result["cell_voltages"] = BMS._cell_voltages(
            self._data, cells=result["cell_count"], start=12, gap=2
        )
        result["temp_values"] = BMS._temp_values(
            self._data,
            values=result["cell_count"],
            start=14,
            gap=2,
            signed=False,
            offset=2731,
            divider=10,
        )
        return result



"""Module to support ANT BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from functools import cache
from typing import Final

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSDp, BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS, barr2str, crc_modbus


class BMS(BaseBMS):
    """ANT BMS implementation."""

    INFO: BMSInfo = {"default_manufacturer": "ANT", "default_model": "smart BMS"}
    _HEAD: Final[bytes] = b"\x7e\xa1"
    _TAIL: Final[bytes] = b"\xaa\x55"
    _MIN_LEN: Final[int] = 10  # frame length without data
    _CMD_STAT: Final[int] = 0x01
    _CMD_DEV: Final[int] = 0x02
    _TEMP_POS: Final[int] = 8
    _MAX_TEMPS: Final[int] = 6
    _CELL_COUNT: Final[int] = 9
    _CELL_POS: Final[int] = 34
    _MAX_CELLS: Final[int] = 32
    _FIELDS: Final[tuple[BMSDp, ...]] = (
        BMSDp("voltage", 38, 2, False, lambda x: x / 100),
        BMSDp("current", 40, 2, True, lambda x: x / 10),
        BMSDp("design_capacity", 50, 4, False, lambda x: x // 1e6),
        BMSDp("battery_level", 42, 2, False, lambda x: x),
        BMSDp(
            "problem_code",
            46,
            2,
            False,
            lambda x: ((x & 0xF00) if (x >> 8) not in (0x1, 0x4, 0xF) else 0)
            | ((x & 0xF) if (x & 0xF) not in (0x1, 0x4, 0xB, 0xF) else 0),
        ),
        BMSDp("cycle_charge", 54, 4, False, lambda x: x / 1e6),
        BMSDp("total_charge", 58, 4, False, lambda x: x // 1000),
        BMSDp("delta_voltage", 82, 2, False, lambda x: x / 1000),
        BMSDp("power", 62, 4, True, lambda x: x / 1),
        BMSDp("chrg_mosfet", 46, 1, False, lambda x: x == 0x1),
        BMSDp("dischrg_mosfet", 47, 1, False, lambda x: x == 0x1),
        BMSDp("balancer", 48, 1, False, lambda x: bool(x & 0x4)),
    )

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Initialize BMS."""
        super().__init__(ble_device, keep_alive)
        self._data_final: bytearray = bytearray()
        self._valid_reply: int = BMS._CMD_STAT | 0x10  # valid reply mask
        self._exp_len: int = 0

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Provide BluetoothMatcher definition."""
        return [
            {
                "local_name": "ANT-BLE*",
                "service_uuid": BMS.uuid_services()[0],
                "manufacturer_id": 0x2313,
                "connectable": True,
            }
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("ffe0")]  # change service UUID here!

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "ffe1"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "ffe1"

    async def _fetch_device_info(self) -> BMSInfo:
        """Fetch the device information via BLE."""

        await self._await_reply(BMS._cmd(BMS._CMD_DEV, 0x026C, 0x20))
        return BMSInfo(
            hw_version=barr2str(self._data_final[6:22]),
            sw_version=barr2str(self._data_final[22:38]),
        )

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {"cycle_capacity", "cycles", "temperature"}
        )  # calculate further values from BMS provided set ones

    async def _init_connection(
        self, char_notify: BleakGATTCharacteristic | int | str | None = None
    ) -> None:
        """Initialize RX/TX characteristics and protocol state."""
        await super()._init_connection(char_notify)
        self._exp_len = 0

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle the RX characteristics notify event (new data arrives)."""

        if (
            data.startswith(BMS._HEAD)
            and len(self._data) >= self._exp_len
            and len(data) >= BMS._MIN_LEN
        ):
            self._data = bytearray()
            self._exp_len = data[5] + BMS._MIN_LEN

        self._data += data
        self._log.debug(
            "RX BLE data (%s): %s", "start" if data == self._data else "cnt.", data
        )

        if len(self._data) < self._exp_len or len(self._data) < BMS._MIN_LEN:
            return

        if self._data[2] != self._valid_reply:
            self._log.debug("unexpected response (type 0x%X)", self._data[2])
            return

        if len(self._data) != self._exp_len and self._data[2] != BMS._CMD_DEV | 0x10:
            # length of CMD_DEV is incorrect, so we ignore the length check here
            self._log.debug(
                "invalid frame length %d != %d", len(self._data), self._exp_len
            )
            return

        if not self._data.endswith(BMS._TAIL):
            self._log.debug("invalid frame end")
            return

        if (crc := crc_modbus(self._data[1 : self._exp_len - 4])) != int.from_bytes(
            self._data[self._exp_len - 4 : self._exp_len - 2], "little"
        ):
            self._log.debug(
                "invalid checksum 0x%X != 0x%X",
                int.from_bytes(
                    self._data[self._exp_len - 4 : self._exp_len - 2], "little"
                ),
                crc,
            )
            return

        self._data_final = self._data.copy()
        self._data_event.set()

    @staticmethod
    @cache
    def _cmd(cmd: int, adr: int, value: int) -> bytes:
        """Assemble a ANT BMS command."""
        frame: bytearray = (
            bytearray([*BMS._HEAD, cmd & 0xFF])
            + adr.to_bytes(2, "little")
            + int.to_bytes(value & 0xFF, 1)
        )
        frame.extend(int.to_bytes(crc_modbus(frame[1:]), 2, "little"))
        return bytes(frame) + BMS._TAIL

    @staticmethod
    def _temp_sensors(data: bytearray, sensors: int, offs: int) -> list[float]:
        return [
            float(int.from_bytes(data[idx : idx + 2], byteorder="little", signed=True))
            for idx in range(offs, offs + sensors * 2, 2)
        ]

    async def _await_reply(
        self,
        data: bytes,
        char: int | str | None = None,
        wait_for_notify: bool = True,
        max_size: int = 0,
    ) -> None:
        """Send data to the BMS and wait for valid reply notification."""

        self._valid_reply = data[2] | 0x10  # expected reply type
        await super()._await_reply(data, char, wait_for_notify, max_size)

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        await self._await_reply(BMS._cmd(BMS._CMD_STAT, 0, 0xBE))

        result: BMSSample = {}
        result["battery_charging"] = self._data_final[7] == 0x2
        result["cell_count"] = min(self._data_final[BMS._CELL_COUNT], BMS._MAX_CELLS)
        result["cell_voltages"] = BMS._cell_voltages(
            self._data_final,
            cells=result["cell_count"],
            start=BMS._CELL_POS,
            byteorder="little",
        )
        result["temp_sensors"] = min(self._data_final[BMS._TEMP_POS], BMS._MAX_TEMPS)
        result["temp_values"] = BMS._temp_sensors(
            self._data_final,
            result["temp_sensors"] + 2,  # + MOSFET, balancer temperature
            BMS._CELL_POS + result["cell_count"] * 2,
        )
        result.update(
            BMS._decode_data(
                BMS._FIELDS,
                self._data_final,
                byteorder="little",
                offset=(result["temp_sensors"] + result["cell_count"]) * 2,
            )
        )

        return result



"""Module to support Offgridtec Smart Pro BMS.

Project: aiobmsble, https://pypi.org/p/aiobmsble/
License: Apache-2.0, http://www.apache.org/licenses/
"""

from collections.abc import Callable
from string import digits, hexdigits
from typing import Any, Final, NamedTuple

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.uuids import normalize_uuid_str

from aiobmsble import BMSInfo, BMSSample, BMSValue, MatcherPattern
from aiobmsble.basebms import BaseBMS


class BMS(BaseBMS):
    """Offgridtec LiFePO4 Smart Pro type A and type B BMS implementation."""

    INFO: BMSInfo = {
        "default_manufacturer": "Offgridtec",
        "default_model": "LiFePo4 Smart Pro",
    }
    _IDX_NAME: Final = 0
    _IDX_LEN: Final = 1
    _IDX_FCT: Final = 2
    # magic crypt sequence of length 16
    _CRYPT_SEQ: Final[list[int]] = [2, 5, 4, 3, 1, 4, 1, 6, 8, 3, 7, 2, 5, 8, 9, 3]

    class _Response(NamedTuple):
        valid: bool
        reg: int
        value: int

    def __init__(self, ble_device: BLEDevice, keep_alive: bool = True) -> None:
        """Intialize private BMS members."""
        super().__init__(ble_device, keep_alive)
        self._type: str = (
            self.name[9]
            if len(self.name) >= 10
            and set(self.name[10:]).issubset(digits)
            else "?"
        )
        self._key: int = (
            sum(
                BMS._CRYPT_SEQ[int(c, 16)]
                for c in (f"{int(self.name[10:]):0>4X}")
            )
            if self._type in "AB"
            else 0
        ) + (5 if (self._type == "A") else 8)
        self._log.info(
            "%s type: %c, ID: %s, key: 0x%X",
            self.bms_id(),
            self._type,
            self.name[10:],
            self._key,
        )
        self._exp_reply: int = 0x0
        self._response: BMS._Response = BMS._Response(False, 0, 0)
        self._REGISTERS: dict[int, tuple[BMSValue, int, Callable[[int], Any]]]
        if self._type == "A":
            self._REGISTERS = {
                # SOC (State of Charge)
                2: ("battery_level", 1, lambda x: x),
                4: ("cycle_charge", 3, lambda x: x / 1000),
                8: ("voltage", 2, lambda x: x / 1000),
                # MOS temperature
                12: ("temperature", 2, lambda x: round(x * 0.1 - 273.15, 1)),
                # 3rd byte of current is 0 (should be 1 as for B version)
                16: ("current", 3, lambda x: x / 100),
                24: ("runtime", 2, lambda x: x * 60),
                44: ("cycles", 2, lambda x: x),
                # Type A batteries have no cell voltage registers
            }
            self._HEADER = "+RAA"
        elif self._type == "B":
            self._REGISTERS = {
                # MOS temperature
                8: ("temperature", 2, lambda x: round(x * 0.1 - 273.15, 1)),
                9: ("voltage", 2, lambda x: x / 1000),
                10: ("current", 3, lambda x: x / 1000),
                # SOC (State of Charge)
                13: ("battery_level", 1, lambda x: x),
                15: ("cycle_charge", 3, lambda x: x / 1000),
                18: ("runtime", 2, lambda x: x * 60),
                23: ("cycles", 2, lambda x: x),
            }
            # add cell voltage registers, note: need to be last!
            self._HEADER = "+R16"
        else:
            self._REGISTERS = {}
            self._log.exception("unkown device type '%c'", self._type)

    @staticmethod
    def matcher_dict_list() -> list[MatcherPattern]:
        """Return a list of Bluetooth matchers."""
        return [
            {
                "local_name": "SmartBat-[AB]*",
                "service_uuid": BMS.uuid_services()[0],
                "connectable": True,
            }
        ]

    @staticmethod
    def uuid_services() -> list[str]:
        """Return list of 128-bit UUIDs of services required by BMS."""
        return [normalize_uuid_str("fff0")]

    @staticmethod
    def uuid_rx() -> str:
        """Return 16-bit UUID of characteristic that provides notification/read property."""
        return "fff4"

    @staticmethod
    def uuid_tx() -> str:
        """Return 16-bit UUID of characteristic that provides write property."""
        return "fff6"

    async def _fetch_device_info(self) -> BMSInfo:  # char "180a" contains garbage
        """Fetch the device information via BLE."""
        return {"serial_number": self.name[10:]}

    @staticmethod
    def _calc_values() -> frozenset[BMSValue]:
        return frozenset(
            {"cycle_capacity", "power", "battery_charging", "delta_voltage"}
        )

    def _notification_handler(
        self, _sender: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        self._log.debug("RX BLE data: %s", data)

        self._response = self._ogt_response(data)

        # check that descrambled message is valid
        if not self._response.valid:
            self._log.debug("response data is invalid")
            return

        if self._response.reg not in (-1, self._exp_reply):
            self._log.debug("wrong register response")
            return

        self._exp_reply = -1
        self._data_event.set()

    def _ogt_response(self, resp: bytearray) -> _Response:
        """Descramble a response from the BMS."""

        try:
            msg: Final[str] = bytearray(
                (resp[x] ^ self._key) for x in range(len(resp))
            ).decode(encoding="ascii")
        except UnicodeDecodeError:
            return BMS._Response(False, -1, 0)

        self._log.debug("response: %s", msg.rstrip("\r\n"))
        # verify correct response
        if len(msg) < 8 or not msg.startswith("+RD,"):
            return BMS._Response(False, -1, 0)
        if msg[4:7] == "Err":
            return BMS._Response(True, -1, 0)
        if not msg.endswith("\r\n") or not all(c in hexdigits for c in msg[4:-2]):
            return BMS._Response(False, -1, 0)

        # 16-bit value in network order (plus optional multiplier for 24-bit values)
        # multiplier has 1 as minimum due to current value in A type battery
        signed: bool = len(msg) > 12
        value: int = int.from_bytes(
            bytes.fromhex(msg[6:10]), byteorder="little", signed=signed
        ) * (max(int(msg[10:12], 16), 1) if signed else 1)
        return BMS._Response(True, int(msg[4:6], 16), value)

    def _ogt_command(self, reg: int, length: int) -> bytes:
        """Put together an scambled query to the BMS."""

        cmd: Final[str] = f"{self._HEADER}{reg:0>2X}{length:0>2X}"
        self._log.debug("command: %s", cmd)

        return bytes(ord(cmd[i]) ^ self._key for i in range(len(cmd)))

    async def _async_update(self) -> BMSSample:
        """Update battery status information."""
        result: BMSSample = {}

        for reg in list(self._REGISTERS):
            self._exp_reply = reg
            await self._await_reply(
                data=self._ogt_command(reg, self._REGISTERS[reg][BMS._IDX_LEN])
            )
            if self._response.reg < 0:
                raise TimeoutError

            name, _length, func = self._REGISTERS[self._response.reg]
            result[name] = func(self._response.value)
            self._log.debug(
                "decoded data: reg: %s (#%i), raw: %i, value: %f",
                name,
                reg,
                self._response.value,
                result.get(name),
            )

        # read cell voltages for type B battery
        if self._type == "B":
            for cell_reg in range(16):
                self._exp_reply = 63 - cell_reg
                await self._await_reply(data=self._ogt_command(63 - cell_reg, 2))
                if self._response.reg < 0:
                    self._log.debug("cell count: %i", cell_reg)
                    break
                result.setdefault("cell_voltages", []).append(
                    self._response.value / 1000
                )

        # remove remaining runtime if battery is charging
        if result.get("runtime") == 0xFFFF * 60:
            del result["runtime"]

        return result
